from typing import Dict, List, Literal, Optional, Text, Union

from openai.types.completion import Completion as OpenaiCompletion
from openai.types.completion_choice import CompletionChoice as OpenaiCompletionChoice
from pydantic import BaseModel, Field


class CompletionRequest(BaseModel):
    model: str
    prompt: Union[Text, List[Union[Text, List[int]]]]
    best_of: Optional[int] = Field(
        default=1,
        description=(
            "Generates best_of completions server-side and returns the 'best' one."
        ),
    )
    echo: Optional[bool] = Field(
        default=False, description="Echo back the prompt in addition to the completion."
    )
    frequency_penalty: Optional[float] = Field(
        default=0,
        ge=-2.0,
        le=2.0,
        description="Penalize new tokens based on their frequency.",
    )
    logit_bias: Optional[Dict[int, int]] = Field(
        default=None,
        description=(
            "Modify the likelihood of specified tokens appearing in the completion."
        ),
    )
    logprobs: Optional[int] = Field(
        default=None,
        ge=0,
        le=5,
        description="Include the log probabilities on the most likely output tokens.",
    )
    max_tokens: Optional[int] = Field(
        default=16,
        description=(
            "The maximum number of tokens that can be generated in the completion."
        ),
    )
    n: Optional[int] = Field(
        default=1, description="How many completions to generate for each prompt."
    )
    presence_penalty: Optional[float] = Field(
        default=0,
        ge=-2.0,
        le=2.0,
        description="Penalize new tokens based on their presence.",
    )
    seed: Optional[int] = Field(
        default=None, description="If specified, sample deterministically."
    )
    stop: Optional[Union[Text, List[Text]]] = Field(
        default=None,
        description="Sequences where the API will stop generating further tokens.",
    )
    stream: Optional[bool] = Field(
        default=False, description="Whether to stream back partial progress."
    )
    suffix: Optional[Text] = Field(
        default=None,
        description="The suffix that comes after a completion of inserted text.",
    )
    temperature: Optional[float] = Field(
        default=1, ge=0, le=2, description="Sampling temperature to use."
    )
    top_p: Optional[float] = Field(
        default=1,
        ge=0,
        le=1,
        description=(
            "An alternative to sampling with temperature, called nucleus sampling."
        ),
    )
    user: Optional[Text] = Field(
        default=None, description="A unique identifier representing your end-user."
    )


class CompletionChoice(OpenaiCompletionChoice):
    finish_reason: Optional[Literal["stop", "length", "content_filter"]] = None


class Completion(OpenaiCompletion):
    choices: List[CompletionChoice]



from typing import TYPE_CHECKING, Literal, TypeVar, overload

import numpy as np
from numpy.typing import NDArray

if TYPE_CHECKING:
    from torch import Tensor

FloatType = TypeVar("FloatType", np.float16, np.float32, np.float64)


def mean_pooling(
    last_hidden_states: NDArray[FloatType], attention_mask: NDArray[FloatType]
) -> NDArray[FloatType]:
    # Convert attention_mask to boolean
    # and expand its dimensions to match last_hidden_states
    attention_mask_expanded = np.expand_dims(attention_mask, axis=-1).astype(bool)
    # Use the attention mask to zero out positions in last_hidden_states
    last_hidden = np.where(attention_mask_expanded, last_hidden_states, 0.0)
    # Sum along the specified axis and compute the average
    sum_hidden = last_hidden.sum(axis=1)
    mask_sum = attention_mask.sum(axis=1, keepdims=True)
    # Avoid division by zero
    mask_sum = np.where(mask_sum == 0, 1, mask_sum)

    average_hidden = sum_hidden / mask_sum
    return average_hidden


@overload
def tensor_to_np(
    tensor: "Tensor", precision: Literal["float16", "half"] = "float16"
) -> NDArray[np.float16]: ...


@overload
def tensor_to_np(
    tensor: "Tensor", precision: Literal["float32", "float"] = "float32"
) -> NDArray[np.float32]: ...


@overload
def tensor_to_np(
    tensor: "Tensor", precision: Literal["float64", "double"] = "float64"
) -> NDArray[np.float64]: ...


def tensor_to_np(
    tensor: "Tensor",
    precision: Literal[
        "float16", "float32", "float64", "half", "float", "double"
    ] = "float32",
) -> NDArray[FloatType]:
    return tensor.detach().cpu().float().numpy()



import time
from typing import Optional, Text

import sqlalchemy as sa
from sqlalchemy.exc import NoResultFound
from sqlalchemy.orm import Session
from yarl import URL

from languru.config import logger
from languru.types.model import Model
from languru.types.model.orm import Base as SQL_Base
from languru.types.model.orm import ModelOrm


class ModelDiscovery:

    @classmethod
    def from_url(cls, url: Text | URL):
        url_str: Text = str(url).strip()
        if (
            url_str.startswith("sqlite")
            or url_str.startswith("postgresql")
            or url_str.startswith("postgres")
            or url_str.startswith("mysql")
        ):
            return SqlModelDiscovery(url)
        else:
            logger.error(f"Unsupported discovery url: {url_str}")
            raise ValueError(f"Unsupported discovery url: {url_str}")

    def touch(self) -> bool:
        raise NotImplementedError  # pragma: no cover

    def register(self, model: Model) -> Model:
        raise NotImplementedError  # pragma: no cover

    def retrieve(self, id: Text) -> Model | None:
        raise NotImplementedError  # pragma: no cover

    def list(
        self,
        id: Optional[Text] = None,
        owned_by: Optional[Text] = None,
        created_from: Optional[int] = None,
        created_to: Optional[int] = None,
        limit: int = 20,
    ) -> list[Model]:
        raise NotImplementedError  # pragma: no cover


class SqlModelDiscovery(ModelDiscovery):
    def __init__(self, url: Text | URL):
        self.url: Text = str(url)
        connect_kwargs = {}
        if self.url.startswith("sqlite"):
            connect_kwargs["check_same_thread"] = False
        self._engine = sa.create_engine(self.url, connect_args=connect_kwargs)

    @property
    def sql_engine(self):
        return self._engine

    def touch(self) -> bool:
        return SQL_Base.metadata.create_all(self.sql_engine)

    def register(self, model: Model, created: int | None = None) -> Model:
        if not isinstance(model, Model):
            raise TypeError(f"Expected Model, got {type(model)}")
        if not model.id:
            raise ValueError("Model id is required")

        created = int(time.time()) if created is None else created
        with Session(self.sql_engine) as session:
            # update or create
            try:
                model_orm = (
                    session.query(ModelOrm)
                    .filter(
                        ModelOrm.id == model.id, ModelOrm.owned_by == model.owned_by
                    )
                    .one()
                )
                model_orm.created = created
            except NoResultFound:
                # create one
                model_orm = ModelOrm(
                    id=model.id,
                    created=created,
                    object=model.object,
                    owned_by=model.owned_by,
                )
                session.add(model_orm)
            session.commit()
            session.refresh(model_orm)

            return Model.model_validate(model_orm)

    def retrieve(self, id: Text) -> Model | None:
        with Session(self.sql_engine) as session:
            model_orm = (
                session.query(ModelOrm)
                .filter(ModelOrm.id == id)
                .order_by(ModelOrm.created.desc())
                .one_or_none()
            )
            if model_orm is None:
                return None
            return Model.model_validate(model_orm)

    def list(
        self,
        id: Optional[Text] = None,
        owned_by: Optional[Text] = None,
        created_from: Optional[int] = None,
        created_to: Optional[int] = None,
        limit: int = 20,
    ) -> list[Model]:
        with Session(self.sql_engine) as session:
            query = session.query(ModelOrm)
            if id:
                query = query.filter(ModelOrm.id == id)
            if owned_by:
                query = query.filter(ModelOrm.owned_by == owned_by)
            if created_from:
                query = query.filter(ModelOrm.created >= created_from)
            if created_to:
                query = query.filter(ModelOrm.created <= created_to)
            query = query.order_by(ModelOrm.created.desc()).limit(limit)
            model_records = query.all()
            return [Model.model_validate(model_orm) for model_orm in model_records]



from typing import TYPE_CHECKING, Generator, List, Optional, Text, Union

import openai

from languru.action.base import ActionBase, ModelDeploy
from languru.config import logger

if TYPE_CHECKING:
    from openai.types import (
        Completion,
        CreateEmbeddingResponse,
        ModerationCreateResponse,
    )
    from openai.types.chat import (
        ChatCompletion,
        ChatCompletionChunk,
        ChatCompletionMessageParam,
    )


class OpenaiAction(ActionBase):
    model_deploys = (
        ModelDeploy("babbage-002", "babbage-002"),
        ModelDeploy("dall-e-2", "dall-e-2"),
        ModelDeploy("dall-e-3", "dall-e-3"),
        ModelDeploy("davinci-002", "davinci-002"),
        ModelDeploy("gpt-3.5-turbo", "gpt-3.5-turbo"),
        ModelDeploy("gpt-3.5-turbo-0125", "gpt-3.5-turbo-0125"),
        ModelDeploy("gpt-3.5-turbo-0301", "gpt-3.5-turbo-0301"),
        ModelDeploy("gpt-3.5-turbo-0613", "gpt-3.5-turbo-0613"),
        ModelDeploy("gpt-3.5-turbo-1106", "gpt-3.5-turbo-1106"),
        ModelDeploy("gpt-3.5-turbo-16k", "gpt-3.5-turbo-16k"),
        ModelDeploy("gpt-3.5-turbo-16k-0613", "gpt-3.5-turbo-16k-0613"),
        ModelDeploy("gpt-3.5-turbo-instruct", "gpt-3.5-turbo-instruct"),
        ModelDeploy("gpt-3.5-turbo-instruct-0914", "gpt-3.5-turbo-instruct-0914"),
        ModelDeploy("gpt-4", "gpt-4"),
        ModelDeploy("gpt-4-0125-preview", "gpt-4-0125-preview"),
        ModelDeploy("gpt-4-0613", "gpt-4-0613"),
        ModelDeploy("gpt-4-1106-preview", "gpt-4-1106-preview"),
        ModelDeploy("gpt-4-turbo-preview", "gpt-4-turbo-preview"),
        ModelDeploy("gpt-4-vision-preview", "gpt-4-vision-preview"),
        ModelDeploy("text-embedding-3-large", "text-embedding-3-large"),
        ModelDeploy("text-embedding-3-small", "text-embedding-3-small"),
        ModelDeploy("text-embedding-ada-002", "text-embedding-ada-002"),
        ModelDeploy("text-moderation-latest", "text-moderation-latest"),
        ModelDeploy("text-moderation-stable", "text-moderation-stable"),
        ModelDeploy("tts-1", "tts-1"),
        ModelDeploy("tts-1-1106", "tts-1-1106"),
        ModelDeploy("tts-1-hd", "tts-1-hd"),
        ModelDeploy("tts-1-hd-1106", "tts-1-hd-1106"),
        ModelDeploy("whisper-1", "whisper-1"),
        ModelDeploy("openai/babbage-002", "babbage-002"),
        ModelDeploy("openai/dall-e-2", "dall-e-2"),
        ModelDeploy("openai/dall-e-3", "dall-e-3"),
        ModelDeploy("openai/davinci-002", "davinci-002"),
        ModelDeploy("openai/gpt-3.5-turbo", "gpt-3.5-turbo"),
        ModelDeploy("openai/gpt-3.5-turbo-0125", "gpt-3.5-turbo-0125"),
        ModelDeploy("openai/gpt-3.5-turbo-0301", "gpt-3.5-turbo-0301"),
        ModelDeploy("openai/gpt-3.5-turbo-0613", "gpt-3.5-turbo-0613"),
        ModelDeploy("openai/gpt-3.5-turbo-1106", "gpt-3.5-turbo-1106"),
        ModelDeploy("openai/gpt-3.5-turbo-16k", "gpt-3.5-turbo-16k"),
        ModelDeploy("openai/gpt-3.5-turbo-16k-0613", "gpt-3.5-turbo-16k-0613"),
        ModelDeploy("openai/gpt-3.5-turbo-instruct", "gpt-3.5-turbo-instruct"),
        ModelDeploy(
            "openai/gpt-3.5-turbo-instruct-0914", "gpt-3.5-turbo-instruct-0914"
        ),
        ModelDeploy("openai/gpt-4", "gpt-4"),
        ModelDeploy("openai/gpt-4-0125-preview", "gpt-4-0125-preview"),
        ModelDeploy("openai/gpt-4-0613", "gpt-4-0613"),
        ModelDeploy("openai/gpt-4-1106-preview", "gpt-4-1106-preview"),
        ModelDeploy("openai/gpt-4-turbo-preview", "gpt-4-turbo-preview"),
        ModelDeploy("openai/gpt-4-vision-preview", "gpt-4-vision-preview"),
        ModelDeploy("openai/text-embedding-3-large", "text-embedding-3-large"),
        ModelDeploy("openai/text-embedding-3-small", "text-embedding-3-small"),
        ModelDeploy("openai/text-embedding-ada-002", "text-embedding-ada-002"),
        ModelDeploy("openai/text-moderation-latest", "text-moderation-latest"),
        ModelDeploy("openai/text-moderation-stable", "text-moderation-stable"),
        ModelDeploy("openai/tts-1", "tts-1"),
        ModelDeploy("openai/tts-1-1106", "tts-1-1106"),
        ModelDeploy("openai/tts-1-hd", "tts-1-hd"),
        ModelDeploy("openai/tts-1-hd-1106", "tts-1-hd-1106"),
        ModelDeploy("openai/whisper-1", "whisper-1"),
    )

    def __init__(
        self,
        *args,
        api_key: Optional[Text] = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        self._client = openai.OpenAI(api_key=api_key)

    def name(self):
        return "openai_action"

    def health(self) -> bool:
        try:
            self._client.models.retrieve(model="gpt-3.5-turbo")
            return True
        except Exception as e:
            logger.error(f"OpenAI health check failed: {e}")
            return False

    def chat(
        self, messages: List["ChatCompletionMessageParam"], *args, model: Text, **kwargs
    ) -> "ChatCompletion":
        model = self.validate_model(model)
        if kwargs.get("frequency_penalty") == 0.0:
            kwargs.pop("frequency_penalty")
        chat_completion = self._client.chat.completions.create(
            messages=messages, model=model, **kwargs
        )
        return chat_completion

    def chat_stream(
        self, messages: List["ChatCompletionMessageParam"], *args, model: Text, **kwargs
    ) -> Generator["ChatCompletionChunk", None, None]:
        if "stream" in kwargs and not kwargs["stream"]:
            logger.warning(f"Chat stream should be True, but got: {kwargs['stream']}")
        model = self.validate_model(model)
        kwargs.pop("stream", None)
        if kwargs.get("frequency_penalty") == 0.0:
            kwargs.pop("frequency_penalty")
        chat_completion_stream = self._client.chat.completions.create(
            messages=messages, model=model, stream=True, **kwargs
        )
        for _chat in chat_completion_stream:
            yield _chat

    def text_completion(
        self, prompt: Text, *args, model: Text, **kwargs
    ) -> "Completion":
        model = self.validate_model(model)
        completion = self._client.completions.create(
            prompt=prompt, model=model, **kwargs
        )
        return completion

    def text_completion_stream(
        self, prompt: Text, *args, model: Text, **kwargs
    ) -> Generator["Completion", None, None]:
        if "stream" in kwargs and not kwargs["stream"]:
            logger.warning(
                f"Text completion stream should be True, but got: {kwargs['stream']}"
            )
        kwargs.pop("stream", None)
        model = self.validate_model(model)
        completion_stream = self._client.completions.create(
            prompt=prompt, model=model, stream=True, **kwargs
        )
        for _completion in completion_stream:
            yield _completion

    def embeddings(
        self, input: Union[Text, List[Text]], *args, model: Text, **kwargs
    ) -> "CreateEmbeddingResponse":
        model = self.validate_model(model)
        embeddings = self._client.embeddings.create(input=input, model=model, **kwargs)
        return embeddings

    def moderations(
        self, input: Text, *args, model: Text, **kwargs
    ) -> "ModerationCreateResponse":
        model = self.validate_model(model)
        moderation = self._client.moderations.create(input=input, model=model, **kwargs)
        return moderation



import logging
import logging.config
import os
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import TYPE_CHECKING, Literal, Optional, Text

import pytz
from colorama import Fore, Style, init
from pydantic_settings import BaseSettings
from rich.console import Console
from rich.table import Table

from languru.version import VERSION

if TYPE_CHECKING:
    from fastapi import FastAPI

console = Console()


# string enum
class AppType(str, Enum):
    llm = "llm"
    agent = "agent"


class ServerBaseSettings(BaseSettings):
    """Settings for the server."""

    # Server
    APP_NAME: Text = "languru"
    SERVICE_NAME: Text = APP_NAME
    APP_VERSION: Text = VERSION
    APP_TYPE: Optional[Literal[AppType.llm, AppType.agent]] = None
    is_production: bool = False
    is_development: bool = True
    is_testing: bool = False
    debug: bool = is_development
    logging_level: Text = "DEBUG"
    logs_dir: Text = "logs"
    HOST: Text = "0.0.0.0"
    DEFAULT_PORT: int
    PORT: Optional[int] = None
    WORKERS: int = 1
    RELOAD: bool = True
    LOG_LEVEL: Text = "debug"
    USE_COLORS: bool = True
    RELOAD_DELAY: float = 5.0
    DATA_DIR: Text = str(Path("./data").absolute())

    # Resources configuration
    openai_available: bool = True if os.environ.get("OPENAI_API_KEY") else False


class LlmSettings(ServerBaseSettings):
    APP_NAME: Text = "languru-llm"
    SERVICE_NAME: Text = APP_NAME
    APP_TYPE: Literal[AppType.llm] = AppType.llm
    DEFAULT_PORT: int = 8682

    # LLM Server Configuration
    LLM_BASE_URL: Text = "http://0.0.0.0:8682/v1"
    ACTION_BASE_URL: Text = LLM_BASE_URL  # Deprecated, use LLM_BASE_URL instead
    ACTION_ENDPOINT_URL: Text = LLM_BASE_URL  # Deprecated, use LLM_BASE_URL instead
    AGENT_BASE_URL: Optional[Text] = None
    MODEL_REGISTER_PERIOD: int = 8
    MODEL_REGISTER_FAIL_PERIOD: int = 60

    # Hardware device configuration
    device: Optional[Text] = None
    dtype: Optional[Text] = None

    # LLM action configuration
    action: Text = "languru.action.openai.OpenaiAction"


class AgentSettings(ServerBaseSettings):
    # Server
    APP_NAME: Text = "languru-agent"
    SERVICE_NAME: Text = APP_NAME
    APP_TYPE: Literal[AppType.agent] = AppType.agent
    DEFAULT_PORT: int = 8680
    DATA_DIR: Text = str(Path("./data").absolute())

    # Model discovery configuration
    MODEL_REGISTER_PERIOD: int = 10
    url_model_discovery: Text = f"sqlite:///{DATA_DIR}/languru_model_discovery.db"


class IsoDatetimeFormatter(logging.Formatter):
    def formatTime(self, record, datefmt=None):
        record_datetime = datetime.fromtimestamp(record.created).astimezone(
            pytz.timezone("Asia/Taipei")
        )
        t = record_datetime.strftime("%Y-%m-%dT%H:%M:%S")
        z = record_datetime.strftime("%z")
        ms_exp = record_datetime.microsecond // 1000
        s = f"{t}.{ms_exp:03d}{z}"
        return s


class ColoredIsoDatetimeFormatter(IsoDatetimeFormatter):
    COLORS = {
        "WARNING": Fore.YELLOW,
        "INFO": Fore.GREEN,
        "DEBUG": Fore.BLUE,
        "CRITICAL": Fore.RED,
        "ERROR": Fore.RED,
    }
    MSG_COLORS = {
        "WARNING": Fore.YELLOW,
        "INFO": Fore.GREEN,
        "CRITICAL": Fore.RED,
        "ERROR": Fore.RED,
    }

    def format(self, record):
        levelname = record.levelname
        if levelname in self.COLORS:
            record.levelname = (
                self.COLORS[levelname] + f"{levelname:8s}" + Style.RESET_ALL
            )
            record.name = Fore.BLUE + record.name + Style.RESET_ALL
            if not isinstance(record.msg, Text):
                record.msg = str(record.msg)
            if levelname in self.MSG_COLORS:
                record.msg = self.COLORS[levelname] + record.msg + Style.RESET_ALL
        return super(ColoredIsoDatetimeFormatter, self).format(record)


def default_logging_config(settings: "ServerBaseSettings"):
    d = {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "colored_formatter": {
                "()": ColoredIsoDatetimeFormatter,
                "format": "%(asctime)s %(levelname)-8s %(name)s  - %(message)s",
            },
            "message_formatter": {"format": "%(message)s"},
            "file_formatter": {
                "()": IsoDatetimeFormatter,
                "format": "%(asctime)s %(levelname)-8s %(name)s  - %(message)s",
            },
        },
        "handlers": {
            "console_handler": {
                "level": "DEBUG",
                "class": "logging.StreamHandler",
                "formatter": "colored_formatter",
            },
            "file_handler": {
                "level": settings.logging_level,
                "class": "logging.handlers.RotatingFileHandler",
                "filename": Path(settings.logs_dir)
                .joinpath(f"{settings.APP_NAME}.log")
                .resolve(),
                "formatter": "file_formatter",
                "maxBytes": 2097152,
                "backupCount": 20,
            },
            "error_handler": {
                "level": "WARNING",
                "class": "logging.FileHandler",
                "filename": Path(settings.logs_dir)
                .joinpath(f"{settings.APP_NAME}.error.log")
                .resolve(),
                "formatter": "file_formatter",
            },
        },
        "loggers": {
            "languru": {
                "level": "DEBUG",
                "handlers": ["file_handler", "error_handler", "console_handler"],
                "propagate": True,
            },
            settings.APP_NAME: {
                "level": "DEBUG",
                "handlers": ["file_handler", "error_handler", "console_handler"],
                "propagate": True,
            },
        },
    }
    return d


def init_logger_config(settings: "ServerBaseSettings") -> None:
    if settings.USE_COLORS:
        init(autoreset=True)
    logging.config.dictConfig(default_logging_config(settings))
    return


def init_paths(settings: "ServerBaseSettings") -> None:
    Path(settings.logs_dir).mkdir(parents=True, exist_ok=True, mode=0o770)
    Path(settings.DATA_DIR).mkdir(parents=True, exist_ok=True, mode=0o770)
    return


def pretty_print_app_routes(app: "FastAPI") -> None:
    """Show all routes in the FastAPI app."""

    table = Table(title="\nLanguru Routes")
    table.add_column("Path", style="cyan", no_wrap=True)
    table.add_column("Name", style="magenta", no_wrap=True)

    path_names = [
        (str(getattr(route, "path", "null")), str(getattr(route, "name", "null")))
        for route in app.routes
    ]
    url_list = [
        {"path": path_name[0], "name": path_name[1]} for path_name in path_names
    ]
    url_list.sort(key=lambda item: item["path"])
    for _url_item in url_list:
        table.add_row(_url_item["path"], _url_item["name"])
    console.print(table)



import logging
from typing import TYPE_CHECKING, Any, Optional, Text, Type, TypeVar

from languru.config import logger as languru_logger

if TYPE_CHECKING:
    from fastapi import FastAPI

T = TypeVar("T")

NOT_SET = object()


def get_value_from_app(
    app: "FastAPI",
    key: Text,
    value_typing: Optional[Type[T]] = None,
    default: Any = NOT_SET,
    logger: Optional[logging.Logger] = None,
) -> T:
    logger = logger or languru_logger
    out = default
    if hasattr(app.state, key):
        if value_typing is None:
            out = getattr(app.state, key)
        else:
            if isinstance(getattr(app.state, key, None), value_typing) is False:
                logger.warning(f"Key {key} not of type {value_typing} in app.state")
            out = getattr(app.state, key)
    elif key in app.extra:
        if value_typing is None:
            out = app.extra[key]
        else:
            if isinstance(app.extra[key], value_typing) is False:
                logger.warning(f"Key {key} not of type {value_typing} in app.extra")
            out = app.extra[key]
    if out is NOT_SET:
        raise ValueError(
            f"Key {key} not found or not of type {value_typing} "
            + "in app.state or app.extra"
        )
    return out



import logging
import math
import time
from typing import Annotated, Any, Dict, List, Optional, Text, cast

from fastapi import APIRouter, Depends, HTTPException
from fastapi import Path as PathParam
from fastapi import Query, Request
from openai.pagination import SyncPage
from openai.types import ModelDeleted
from pyassorted.asyncio.executor import run_func

from languru.exceptions import ModelNotFound
from languru.server.config import (
    AgentSettings,
    AppType,
    LlmSettings,
    ServerBaseSettings,
)
from languru.server.deps.common import app_settings
from languru.server.utils.common import get_value_from_app
from languru.types.model import Model

router = APIRouter()


class ModelsHandler:
    async def handle_models_request(
        self,
        request: "Request",
        id: Optional[Text],
        owned_by: Optional[Text],
        created_from: Optional[int],
        created_to: Optional[int],
        settings: "ServerBaseSettings",
    ) -> "SyncPage[Model]":
        if settings.APP_TYPE == AppType.agent:
            settings = cast(AgentSettings, settings)
            return await self.handle_models_agent(
                request=request,
                id=id,
                owned_by=owned_by,
                created_from=created_from,
                created_to=created_to,
                settings=settings,
            )
        if settings.APP_TYPE == AppType.llm:
            settings = cast(LlmSettings, settings)
            return await self.handle_models_llm(
                request=request,
                id=id,
                owned_by=owned_by,
                created_from=created_from,
                created_to=created_to,
                settings=settings,
            )

        # Not implemented or unknown app server type
        raise HTTPException(
            status_code=500,
            detail=(
                f"Unknown app server type: {settings.APP_TYPE}"
                if settings.APP_TYPE
                else "App server type not implemented"
            ),
        )

    async def handle_models_agent(
        self,
        request: "Request",
        id: Optional[Text],
        owned_by: Optional[Text],
        created_from: Optional[int],
        created_to: Optional[int],
        settings: "AgentSettings",
    ) -> "SyncPage[Model]":
        from languru.resources.model.discovery import ModelDiscovery

        model_discovery: "ModelDiscovery" = get_value_from_app(
            request.app, key="model_discovery", value_typing=ModelDiscovery
        )

        if created_from is None and created_to is not None:
            created_from = math.floor(
                time.time() - float(settings.MODEL_REGISTER_PERIOD)
            )
        retrieved_models = await run_func(
            model_discovery.list,
            id=id,
            owned_by=owned_by,
            created_from=created_from,
            created_to=created_to,
        )
        return SyncPage(data=retrieved_models, object="list")

    async def handle_models_llm(
        self,
        request: "Request",
        id: Optional[Text],
        owned_by: Optional[Text],
        created_from: Optional[int],
        created_to: Optional[int],
        settings: "LlmSettings",
    ) -> "SyncPage[Model]":
        from languru.action.base import ActionBase

        action: "ActionBase" = get_value_from_app(
            request.app, key="action", value_typing=ActionBase
        )
        models: List["Model"] = []
        if action.model_deploys:
            for model_deploy in action.model_deploys:
                models.append(
                    Model(
                        id=model_deploy.model_deploy_name,
                        created=int(time.time()),
                        object="model",
                        owned_by=settings.LLM_BASE_URL,
                    )
                )
        return SyncPage(data=models, object="list")


class RetrieveModelHandler:
    async def handle_model_request(
        self,
        request: "Request",
        model: Text,
        settings: "ServerBaseSettings",
    ) -> Model:
        if settings.APP_TYPE == AppType.agent:
            settings = cast(AgentSettings, settings)
            return await self.handle_model_agent(
                request=request, model=model, settings=settings
            )
        if settings.APP_TYPE == AppType.llm:
            settings = cast(LlmSettings, settings)
            return await self.handle_model_llm(
                request=request, model=model, settings=settings
            )

        # Not implemented or unknown app server type
        raise HTTPException(
            status_code=500,
            detail=(
                f"Unknown app server type: {settings.APP_TYPE}"
                if settings.APP_TYPE
                else "App server type not implemented"
            ),
        )

    async def handle_model_agent(
        self,
        request: "Request",
        model: Text,
        settings: "AgentSettings",
    ) -> Model:
        from languru.resources.model.discovery import ModelDiscovery

        model_discovery: "ModelDiscovery" = get_value_from_app(
            request.app, key="model_discovery", value_typing=ModelDiscovery
        )
        retrieved_model = await run_func(model_discovery.retrieve, id=model)
        if retrieved_model is None:
            raise HTTPException(status_code=404, detail="Model not found")
        return retrieved_model

    async def handle_model_llm(
        self,
        request: "Request",
        model: Text,
        settings: "LlmSettings",
    ) -> Model:
        from languru.action.base import ActionBase

        action: "ActionBase" = get_value_from_app(
            request.app, key="action", value_typing=ActionBase
        )
        logger = get_value_from_app(
            request.app,
            key="logger",
            value_typing=logging.Logger,
            default=logging.getLogger(settings.APP_NAME),
        )
        try:
            model = action.get_model_name(model)
        except ModelNotFound as e:
            logger.debug(f"Model not found: {e}")
            raise HTTPException(status_code=404, detail=str(e))
        return Model(
            id=model,
            created=int(time.time()),
            object="model",
            owned_by=settings.LLM_BASE_URL,
        )


class ModelRegisterHandler:
    async def handle_model_register_request(
        self, request: "Request", model: Model, settings: "ServerBaseSettings"
    ) -> Dict[Text, Any]:
        if settings.APP_TYPE == AppType.agent:
            settings = cast(AgentSettings, settings)
            return await self.handle_model_register_agent(
                request=request, model=model, settings=settings
            )
        if settings.APP_TYPE == AppType.llm:
            raise HTTPException(
                status_code=403,
                detail="Model registration not allowed for this app server type",
            )

        # Not implemented or unknown app server type
        raise HTTPException(
            status_code=500,
            detail=(
                f"Unknown app server type: {settings.APP_TYPE}"
                if settings.APP_TYPE
                else "App server type not implemented"
            ),
        )

    async def handle_model_register_agent(
        self, request: "Request", model: Model, settings: "AgentSettings"
    ) -> Dict[Text, Any]:
        from languru.resources.model.discovery import ModelDiscovery

        model_discovery: "ModelDiscovery" = get_value_from_app(
            request.app, key="model_discovery", value_typing=ModelDiscovery
        )
        logger = get_value_from_app(
            request.app, key="logger", value_typing=logging.Logger
        )
        try:
            await run_func(model_discovery.register, model)
            return {"acknowledge": True}
        except Exception as e:
            logger.exception(e)
            logger.error(f"Failed to register model: {e}")
            raise HTTPException(status_code=500, detail="Failed to register model")


@router.get("/models", summary="List models")
async def get_models(
    request: Request,
    id: Annotated[Text | None, "Model ID"] = Query(None),
    owned_by: Annotated[Text | None, "Model owned by"] = Query(None),
    created_from: Annotated[int | None, "Model created from timestamp"] = Query(None),
    created_to: Annotated[int | None, "Model created to timestamp"] = Query(None),
    settings: ServerBaseSettings = Depends(app_settings),
) -> SyncPage[Model]:
    return await ModelsHandler().handle_models_request(
        request=request,
        id=id,
        owned_by=owned_by,
        created_from=created_from,
        created_to=created_to,
        settings=settings,
    )


@router.get("/models/{model}", summary="Retrieve model")
async def get_model(
    request: Request,
    model: Text = PathParam(...),
    settings: ServerBaseSettings = Depends(app_settings),
) -> Model:
    return await RetrieveModelHandler().handle_model_request(
        request=request, model=model, settings=settings
    )


@router.delete("/models/{model}", summary="Delete a fine-tuned model")
async def delete_model(request: Request, model: Text = PathParam(...)) -> ModelDeleted:
    # Not implemented
    return ModelDeleted(id=model, deleted=True, object="model")


@router.post("/models/register", summary="Register models")
async def register_models(
    request: Request,
    model: Model,
    settings: ServerBaseSettings = Depends(app_settings),
):
    return await ModelRegisterHandler().handle_model_register_request(
        request=request, model=model, settings=settings
    )



from typing import Optional, Text

from languru.config import logger

# Support dtype
cpu_supported_dtype = ("float32", "float64")
mps_supported_dtype = ("float16", "float32", "float64", "bfloat16")


def validate_device(device: Optional[Text] = None) -> Text:
    if device is not None and device.strip().lower() == "auto":
        return "auto"
    if device is not None and device.strip().lower() == "cpu":
        return "cpu"
    try:
        import torch
    except ImportError:
        logger.warning("Torch is not installed. The device will be CPU.")
        return "cpu"
    # Try to find the device
    if device is None:
        if torch.cuda.is_available():
            logger.debug("Device found: CUDA")
            return "cuda"
        elif torch.backends.mps.is_available():
            logger.debug("Device found: MPS")
            return "mps"
        else:
            logger.debug("Device found: CPU")
            return "cpu"
    # Validate cuda device
    if device.strip().startswith("cuda"):
        if not torch.cuda.is_available():
            logger.warning("CUDA is not available. The device will be CPU.")
            return "cpu"
        return device
    # Validate mps device
    if device.strip().startswith("mps"):
        if not torch.backends.mps.is_available():
            logger.warning("MPS is not available. The device will be CPU.")
            return "cpu"
        return device
    # Validate cpu device
    if device.strip() == "cpu":
        return device
    # Validate unknown device
    logger.warning(f"Unknown device: {device}. The device will be CPU.")
    return "cpu"


def validate_dtype(device: Text, dtype: Optional[Text] = None):
    if dtype is None:
        return
    device_str_value = device.strip().lower()
    dtype_value = dtype.strip().lower()

    if device_str_value == "cpu":
        if dtype_value not in cpu_supported_dtype:
            raise ValueError(f"Invalid dtype for CPU: {dtype}")
    elif device_str_value.startswith("mps"):
        if dtype_value not in mps_supported_dtype:
            raise ValueError(f"Invalid dtype for MPS: {dtype}")
    elif device_str_value.startswith("auto"):
        return
    elif device_str_value.startswith("cuda"):
        return
    else:
        logger.debug(f"Unhandled device and dtype yet: {device} {dtype}.")



from fastapi import APIRouter

from languru.server.api.v1.chat import router as chat_router
from languru.server.api.v1.completions import router as completions_router
from languru.server.api.v1.embeddings import router as embeddings_router
from languru.server.api.v1.model import router as model_router
from languru.server.api.v1.moderations import router as moderations_router

router = APIRouter()


router.include_router(router=model_router, tags=["model"])
router.include_router(router=chat_router, tags=["chat"])
router.include_router(router=completions_router, tags=["completions"])
router.include_router(router=embeddings_router, tags=["embeddings"])
router.include_router(router=moderations_router, tags=["moderations"])



import importlib
import json
import time
from typing import TYPE_CHECKING
from unittest.mock import MagicMock, patch

import pytest
from fastapi.testclient import TestClient

import languru.server.main
from languru.server.config import AppType

if TYPE_CHECKING:
    from _pytest.monkeypatch import MonkeyPatch


@pytest.fixture
def llm_env(monkeypatch: "MonkeyPatch"):
    monkeypatch.setenv("APP_TYPE", AppType.llm)


@pytest.fixture
def agent_env(monkeypatch: "MonkeyPatch"):
    monkeypatch.setenv("APP_TYPE", AppType.agent)


@pytest.fixture
def mocked_openai_text_completion_create():
    from openai.resources.completions import Completions as OpenaiCompletions

    from languru.examples.return_values._openai import return_text_completion

    with patch.object(
        OpenaiCompletions,
        "create",
        MagicMock(return_value=return_text_completion),
    ):
        yield


@pytest.fixture
def mocked_openai_text_completion_create_stream():
    from openai.resources.completions import Completions as OpenaiCompletions

    from languru.examples.return_values._openai import return_text_completion_stream

    with patch.object(
        OpenaiCompletions,
        "create",
        MagicMock(return_value=return_text_completion_stream),
    ):
        yield


@pytest.fixture
def mocked_model_discovery_list():
    from languru.resources.model.discovery import ModelDiscovery, SqlModelDiscovery
    from languru.types.model import Model

    return_model_discovery_list = [
        Model(
            id="gpt-3.5-turbo",
            created=int(time.time()) - 1,
            object="model",
            owned_by="http://0.0.0.0:8682/v1",
        )
    ]
    with patch.object(
        ModelDiscovery, "list", MagicMock(return_value=return_model_discovery_list)
    ), patch.object(
        SqlModelDiscovery, "list", MagicMock(return_value=return_model_discovery_list)
    ):
        yield


def test_llm_app_text_completions(llm_env, mocked_openai_text_completion_create):
    importlib.reload(languru.server.main)

    with TestClient(languru.server.main.app) as client:
        completion_call = {
            "model": "gpt-3.5-turbo-instruct",
            "prompt": "Say this is a test",
            "max_tokens": 7,
            "temperature": 0,
        }
        response = client.post("/v1/completions", json=completion_call)
        assert response.status_code == 200
        assert response.json()["choices"][0]["text"]


def test_llm_app_text_completions_stream(
    llm_env, mocked_openai_text_completion_create_stream
):
    importlib.reload(languru.server.main)

    with TestClient(languru.server.main.app) as client:
        completion_call = {
            "model": "gpt-3.5-turbo-instruct",
            "prompt": "Say this is a test",
            "max_tokens": 7,
            "temperature": 0,
            "stream": True,
        }
        with client.stream(
            "POST", url="/v1/completions", json=completion_call
        ) as response:
            answer = ""
            for line in response.iter_lines():
                line = line.replace("data:", "", 1).strip()
                if line and line != "[DONE]":
                    chat_chunk = json.loads(line)
                    if chat_chunk["choices"][0]["text"]:
                        answer += chat_chunk["choices"][0]["text"]
            assert answer


def test_agent_app_chat(
    agent_env, mocked_model_discovery_list, mocked_openai_text_completion_create
):
    importlib.reload(languru.server.main)

    with TestClient(languru.server.main.app) as client:
        completion_call = {
            "model": "gpt-3.5-turbo-instruct",
            "prompt": "Say this is a test",
            "max_tokens": 7,
            "temperature": 0,
        }
        response = client.post("/v1/completions", json=completion_call)
        assert response.status_code == 200


def test_agent_app_chat_stream(
    agent_env, mocked_model_discovery_list, mocked_openai_text_completion_create_stream
):
    importlib.reload(languru.server.main)

    with TestClient(languru.server.main.app) as client:
        completion_call = {
            "model": "gpt-3.5-turbo-instruct",
            "prompt": "Say this is a test",
            "max_tokens": 7,
            "temperature": 0,
            "stream": True,
        }
        with client.stream(
            "POST", url="/v1/completions", json=completion_call
        ) as response:
            answer = ""
            for line in response.iter_lines():
                line = line.replace("data:", "", 1).strip()
                if line and line != "[DONE]":
                    chat_chunk = json.loads(line)
                    if chat_chunk["choices"][0]["text"]:
                        answer += chat_chunk["choices"][0]["text"]
            assert answer



import importlib
import logging
from typing import TYPE_CHECKING, Text, Type

from languru.config import logger

if TYPE_CHECKING:
    from languru.action.base import ActionBase


def load_action(
    action: Text, *, logger: "logging.Logger" = logger, **kwargs
) -> "ActionBase":
    from languru.action.base import ActionBase

    logger.info(f"Loading action class '{action}'")
    if ":" in action:  # action is in the form of module_path:ClassName
        action_module_path, action_class_name = action.rsplit(":", 1)
    else:  # action is in the form of module_path.ClassName
        action_module_path, action_class_name = action.rsplit(".", 1)
    action_cls: Type["ActionBase"] = getattr(
        importlib.import_module(action_module_path), action_class_name
    )
    if issubclass(action_cls, ActionBase) is False:
        raise ValueError(f"Action class '{action}' is not a subclass of ActionBase")
    action_instance = action_cls(**kwargs)
    if action_instance.model_deploys is None:
        raise ValueError(f"Action '{action}' model_deploys is not defined")
    return action_instance



from typing import Dict, List, Optional, Text, Union

from pydantic import BaseModel, Field


class LogitBias(BaseModel):
    # Assuming token ID as integer and bias value as float, adjust types if needed
    token_id: int = Field(..., description="Token ID")
    bias: float = Field(..., ge=-100, le=100, description="Bias value from -100 to 100")


class ResponseFormat(BaseModel):
    type: Text = Field(
        ...,
        description='Type of response format, e.g., "json_object" for JSON mode',
    )


class FunctionChoice(BaseModel):
    type: Text = Field(..., description="Specifies the type, e.g., 'function'")
    function: Optional[Dict[Text, Text]] = None


class Message(BaseModel):
    role: Text
    content: Text


class ChatCompletionRequest(BaseModel):
    messages: List[Message] = Field(
        ..., description="A list of messages comprising the conversation so far"
    )
    model: Text = Field(..., description="ID of the model to use")
    frequency_penalty: Optional[float] = Field(
        0, description="Penalize frequencies of new tokens"
    )
    logit_bias: Optional[Dict[int, float]] = Field(
        None, description="Modify the likelihood of specified tokens"
    )
    logprobs: Optional[bool] = Field(
        False, description="Whether to return log probabilities or not"
    )
    top_logprobs: Optional[int] = Field(
        None, ge=0, le=5, description="Number of most likely tokens to return"
    )
    max_tokens: Optional[int] = Field(
        None, description="Maximum number of tokens that can be generated"
    )
    n: Optional[int] = Field(
        1, description="Number of chat completion choices to generate"
    )
    presence_penalty: Optional[float] = Field(
        0, description="Penalize new tokens based on their presence"
    )
    response_format: Optional[ResponseFormat] = None
    seed: Optional[int] = Field(None, description="Seed for deterministic sampling")
    stop: Optional[Union[Text, List[Text]]] = Field(
        None, description="Sequences where the API will stop generating tokens"
    )
    stream: Optional[bool] = Field(
        False, description="If true, partial message deltas will be sent"
    )
    temperature: Optional[float] = Field(
        1, ge=0, le=2, description="Sampling temperature"
    )
    top_p: Optional[float] = Field(
        1, description="Top p probability mass for nucleus sampling"
    )
    tools: Optional[List[Text]] = Field(
        None, description="List of tools the model may call"
    )
    tool_choice: Optional[Union[Text, FunctionChoice]] = Field(
        None, description="Controls which function is called by the model"
    )
    user: Optional[Text] = Field(
        None, description="A unique identifier representing your end-user"
    )



from typing import TYPE_CHECKING

from fastapi import Request

if TYPE_CHECKING:
    from languru.server.config import ServerBaseSettings


def app_settings(request: "Request") -> "ServerBaseSettings":
    from languru.server.config import ServerBaseSettings
    from languru.server.utils.common import get_value_from_app

    return get_value_from_app(
        request.app, key="settings", value_typing=ServerBaseSettings
    )



import json
import logging
from typing import Any, Dict, Generator, Iterable, List, Literal, Optional, Text, Union

import httpx
from pydantic import BaseModel

from languru.config import logger as languru_logger


async def requests_stream_lines(
    url: Text,
    data: Optional[Dict[Text, Any]] = None,
    method: Literal["POST", "GET"] = "POST",
    headers: Optional[Dict[Text, Text]] = None,
):
    async with httpx.AsyncClient(timeout=30) as client:
        async with client.stream(
            method, str(url), json=data, headers=headers or None
        ) as response:
            async for chunk in response.aiter_lines():
                yield chunk + "\n"


def simple_sse_encode(
    stream: Union[
        Iterable[Union[Text, Dict, List[Any], BaseModel]],
        Generator[Union[Text, Dict, List[Any], BaseModel], None, None],
    ],
    logger: Optional[Union[Text, "logging.Logger"]] = None,
) -> Generator[Text, None, None]:
    logger = logger or languru_logger
    logger = logging.getLogger(logger) if isinstance(logger, Text) else logger
    has_warned = False
    for item in stream:
        if isinstance(item, BaseModel):
            yield f"data: {item.model_dump_json()}\n\n"
        elif isinstance(item, (Dict, List)):
            yield f"data: {json.dumps(item)}\n\n"
        elif isinstance(item, Text):
            yield f"data: {item}\n\n"
        else:
            if has_warned is False:
                logger.warning(
                    f"Unknown type {type(item)} in stream, using str() to encode."
                )
                has_warned = True
            yield f"data: {str(item)}\n\n"
    yield "data: [DONE]\n\n"



import sqlalchemy as sa
from sqlalchemy.orm import declarative_base

tb_name = "models"

Base = declarative_base()


class ModelOrm(Base):
    __tablename__ = tb_name

    db_id = sa.Column(sa.Integer, primary_key=True, autoincrement=True)
    id = sa.Column(sa.String, index=True)
    created = sa.Column(sa.Integer, index=True)
    object = sa.Column(sa.String, default="model")
    owned_by = sa.Column(sa.String, index=True)


__all__ = ["ModelOrm"]



import os
from typing import Optional, Text

import click


@click.group()
def app():
    pass


@click.group()
def agent():
    pass


@click.group()
def llm():
    pass


@click.command("version")
@click.option("--short", "-s", default=False, help="Action to run", is_flag=True)
def pkg_version(short: bool = False):
    from languru.version import VERSION

    if short is True:
        click.echo(VERSION)
    else:
        click.echo(f"languru {VERSION}")


@click.command("run")
@click.option("--port", "-p", default=None, help="Port to run the server")
def agent_run(port: Optional[int]):
    from languru.server.config import AgentSettings, AppType
    from languru.server.main import run_app

    settings = AgentSettings()
    settings.APP_TYPE = os.environ["APP_TYPE"] = AppType.agent

    # Parse port parameter
    if port is None and settings.PORT is not None:
        port = settings.PORT
    elif port is None:
        port = settings.DEFAULT_PORT
    port = int(port)
    if port == settings.DEFAULT_PORT:
        click.echo(f"Using default port {settings.DEFAULT_PORT}")
    else:
        click.echo(f"Using port {port} instead of default port {settings.DEFAULT_PORT}")
    settings.PORT, os.environ["PORT"] = port, str(port)

    click.echo("Running agent server")
    run_app(settings)


@click.command("run")
@click.option("--action", default=None, help="Action to run")
@click.option("--port", "-p", default=None, help="Port to run the server")
@click.option("--auto-port", default=True, help="Port to run the server")
@click.option("--agent-base-url", default=None, help="Agent base URL")
def llm_run(
    action: Optional[Text],
    port: Optional[int],
    auto_port: bool = True,
    agent_base_url: Optional[Text] = None,
):
    from languru.server.config import AppType, LlmSettings
    from languru.server.main import run_app
    from languru.utils.socket import check_port, get_available_port

    settings = LlmSettings()
    settings.APP_TYPE = os.environ["APP_TYPE"] = AppType.llm

    # Parse action parameter
    if action is not None and action.strip() != "":
        os.environ["ACTION"] = settings.action = action.strip()

    # Parse port parameter
    if port is None and settings.PORT is not None:
        port = settings.PORT
    elif port is None and auto_port is True:
        port = get_available_port(settings.DEFAULT_PORT)  # Search for available port
    elif port is None:
        port = settings.DEFAULT_PORT
    int(port)  # Check if port is a valid integer
    if check_port(port) is False:
        raise ValueError(f"The port '{settings.PORT}' is already in use")
    if port == settings.DEFAULT_PORT:
        click.echo(f"Using default port {settings.DEFAULT_PORT}")
    else:
        click.echo(f"Using port {port} instead of default port {settings.DEFAULT_PORT}")
    settings.PORT, os.environ["PORT"] = port, str(port)

    # Parse agent base URL parameter
    if agent_base_url is not None and agent_base_url.strip() != "":
        os.environ["AGENT_BASE_URL"] = settings.AGENT_BASE_URL = agent_base_url.strip()

    # Run the app
    click.echo("Running llm server")
    run_app(settings=settings)


agent.add_command(agent_run)

llm.add_command(llm_run)

app.add_command(agent)
app.add_command(llm)
app.add_command(pkg_version)


if __name__ == "__main__":
    app()



import logging

from pydantic_settings import BaseSettings


class Settings(BaseSettings):
    logger_name: str = "languru"
    debug: bool = True


settings = Settings()

logger = logging.getLogger(settings.logger_name)



import os
from typing import Optional, Sequence, Text

import openai

from languru.action.base import ModelDeploy
from languru.action.openai import OpenaiAction
from languru.config import logger


class PerplexityAction(OpenaiAction):
    model_deploys = (
        ModelDeploy("codellama-34b-instruct", "codellama-34b-instruct"),
        ModelDeploy("codellama-70b-instruct", "codellama-70b-instruct"),
        ModelDeploy("llama-2-70b-chat", "llama-2-70b-chat"),
        ModelDeploy("mistral-7b-instruct", "mistral-7b-instruct"),
        ModelDeploy("mixtral-8x7b-instruct", "mixtral-8x7b-instruct"),
        ModelDeploy("pplx-70b-chat", "pplx-70b-chat"),
        ModelDeploy("pplx-70b-online", "pplx-70b-online"),
        ModelDeploy("pplx-7b-chat", "pplx-7b-chat"),
        ModelDeploy("pplx-7b-online", "pplx-7b-online"),
        ModelDeploy("pplx/codellama-34b-instruct", "codellama-34b-instruct"),
        ModelDeploy("pplx/codellama-70b-instruct", "codellama-70b-instruct"),
        ModelDeploy("pplx/llama-2-70b-chat", "llama-2-70b-chat"),
        ModelDeploy("pplx/mistral-7b-instruct", "mistral-7b-instruct"),
        ModelDeploy("pplx/mixtral-8x7b-instruct", "mixtral-8x7b-instruct"),
        ModelDeploy("pplx/pplx-70b-chat", "pplx-70b-chat"),
        ModelDeploy("pplx/pplx-70b-online", "pplx-70b-online"),
        ModelDeploy("pplx/pplx-7b-chat", "pplx-7b-chat"),
        ModelDeploy("pplx/pplx-7b-online", "pplx-7b-online"),
        ModelDeploy("pplx/sonar-medium-chat", "sonar-medium-chat"),
        ModelDeploy("pplx/sonar-medium-online", "sonar-medium-online"),
        ModelDeploy("pplx/sonar-small-chat", "sonar-small-chat"),
        ModelDeploy("pplx/sonar-small-online", "sonar-small-online"),
        ModelDeploy("sonar-medium-chat", "sonar-medium-chat"),
        ModelDeploy("sonar-medium-online", "sonar-medium-online"),
        ModelDeploy("sonar-small-chat", "sonar-small-chat"),
        ModelDeploy("sonar-small-online", "sonar-small-online"),
    )

    def __init__(
        self,
        *args,
        api_key: Optional[Text] = None,
        base_url: Text = "https://api.perplexity.ai",
        model_deploys: Optional[Sequence[ModelDeploy]] = None,
        **kwargs,
    ):
        self._model_deploys = (
            self.model_deploys if model_deploys is None else model_deploys
        ) or ()

        api_key = (
            api_key
            or os.getenv("PERPLEXITY_API_KEY")
            or os.getenv("PPLX_API_KEY")
            or os.getenv("OPENAI_API_KEY")
        ) or None
        if api_key is None:
            raise ValueError("Perplexity api_key is required")
        elif api_key.startswith("pplx-") is False:
            logger.warning(
                "Perplexity api_key should start with 'pplx-', but got %s", api_key
            )
        self._client = openai.OpenAI(api_key=api_key, base_url=base_url)

    def name(self):
        return "perplexity_action"

    def health(self) -> bool:
        return True  # Wait Perplexity API update



from typing import Literal, Optional, Sequence, Text, Union, cast

import torch
from transformers import PreTrainedTokenizer, PreTrainedTokenizerFast, StoppingCriteria


def remove_special_tokens(
    text: Text, tokenizer: Union[PreTrainedTokenizer, PreTrainedTokenizerFast]
) -> Text:
    """Remove special tokens from the text.

    Parameters
    ----------
    text: Text
        The text to remove special tokens from.

    Returns
    -------
    Text
        The text with special tokens removed.
    """

    return tokenizer.decode(tokenizer.encode(text), skip_special_tokens=True)


class StopAtWordsStoppingCriteria(StoppingCriteria):

    def __init__(self, stop_words_ids: Sequence["torch.LongTensor"]):
        """Create a stopping criteria that stops when the model generates a
        sequence of tokens that matches any of the stop words.

        Parameters
        ----------
        stop_words_ids: Sequence["torch.LongTensor"]
            A sequence of token ids to stop at. e.g.:
            [tensor(50256), tensor([  58,   47, 2885,   60])]
        """

        super().__init__()

        self.stop_words_ids = [self.to_cpu_long(s) for s in stop_words_ids]
        self.stop_reason: Optional[Literal["stop", "length", "content_filter"]] = None

    def __call__(self, input_ids: "torch.LongTensor", scores: "torch.FloatTensor"):
        input_ids_cpu = self.to_cpu_long(input_ids)
        for stop_words_id in self.stop_words_ids:
            if torch.all(
                (stop_words_id == input_ids_cpu[0][-len(stop_words_id) :])
            ).item():
                self.stop_reason = "stop"
                return True
        return False

    def get_stop_reason(self) -> Optional[Literal["stop", "length", "content_filter"]]:
        return self.stop_reason

    def to_cpu_long(self, tensor: "torch.Tensor") -> "torch.LongTensor":
        long_tensor = tensor.cpu().long()
        long_tensor = cast(torch.LongTensor, long_tensor)
        return long_tensor

    @classmethod
    def from_stop_words(
        cls,
        stop_words: Sequence[Text],
        tokenizer: Union["PreTrainedTokenizer", "PreTrainedTokenizerFast"],
    ) -> "StopAtWordsStoppingCriteria":
        """Create a stopping criteria from a sequence of stop words.

        Parameters
        ----------
        stop_words: Sequence[Text]
            A sequence of stop words. e.g.: ["[END]", "[STOP]"
        tokenizer: PreTrainedTokenizer | PreTrainedTokenizerFast
            A tokenizer to convert the stop words to token ids.

        Returns
        -------
        StopAtWordsStoppingCriteria
            A stopping criteria that stops when the model generates
            a sequence of tokens that matches any of the stop words.
        """

        if not stop_words:
            raise ValueError("The `stop_words` cannot be empty")

        stop_words_ids = []
        for stop_word in stop_words:
            _input_ids = tokenizer(stop_word, return_tensors="pt")["input_ids"]
            _input_ids = cast(torch.LongTensor, _input_ids)
            each_stop_words_ids = _input_ids.squeeze()
            stop_words_ids.append(each_stop_words_ids)

        return cls(stop_words_ids=stop_words_ids)



import importlib
import time
from typing import TYPE_CHECKING
from unittest.mock import MagicMock, patch

import pytest
from fastapi.testclient import TestClient

import languru.server.main
from languru.server.config import AppType

if TYPE_CHECKING:
    from _pytest.monkeypatch import MonkeyPatch


@pytest.fixture
def llm_env(monkeypatch: "MonkeyPatch"):
    monkeypatch.setenv("APP_TYPE", AppType.llm)


@pytest.fixture
def agent_env(monkeypatch: "MonkeyPatch"):
    monkeypatch.setenv("APP_TYPE", AppType.agent)


@pytest.fixture
def mocked_openai_embeddings_create():
    from openai.resources.embeddings import Embeddings as OpenaiEmbeddings

    from languru.examples.return_values._openai import return_embedding

    with patch.object(
        OpenaiEmbeddings,
        "create",
        MagicMock(return_value=return_embedding),
    ):
        yield


@pytest.fixture
def mocked_model_discovery_list():
    from languru.resources.model.discovery import ModelDiscovery, SqlModelDiscovery
    from languru.types.model import Model

    return_model_discovery_list = [
        Model(
            id="gpt-3.5-turbo",
            created=int(time.time()) - 1,
            object="model",
            owned_by="http://0.0.0.0:8682/v1",
        )
    ]
    with patch.object(
        ModelDiscovery, "list", MagicMock(return_value=return_model_discovery_list)
    ), patch.object(
        SqlModelDiscovery, "list", MagicMock(return_value=return_model_discovery_list)
    ):
        yield


def test_llm_app_embedding(llm_env):
    importlib.reload(languru.server.main)

    with TestClient(languru.server.main.app) as client:
        embedding_call = {
            "input": ["Hello", "world!"],
            "model": "text-embedding-ada-002",
            "encoding_format": "float",
        }
        response = client.post("/v1/embeddings", json=embedding_call)
        assert response.status_code == 200
        assert len(response.json()["data"]) == len(embedding_call["input"])


def test_agent_app_embedding(
    agent_env, mocked_model_discovery_list, mocked_openai_embeddings_create
):
    importlib.reload(languru.server.main)

    with TestClient(languru.server.main.app) as client:
        embedding_call = {
            "input": ["Hello", "world!"],
            "model": "text-embedding-ada-002",
            "encoding_format": "float",
        }
        response = client.post("/v1/embeddings", json=embedding_call)
        assert response.status_code == 200



from functools import lru_cache
from typing import (
    TYPE_CHECKING,
    Generator,
    List,
    NamedTuple,
    Optional,
    Sequence,
    Text,
    Union,
)

from languru.exceptions import ModelNotFound
from languru.utils.common import str_strong_casefold

if TYPE_CHECKING:
    from openai.types import (
        Completion,
        CreateEmbeddingResponse,
        ModerationCreateResponse,
    )
    from openai.types.chat import (
        ChatCompletion,
        ChatCompletionChunk,
        ChatCompletionMessageParam,
    )


ModelDeploy = NamedTuple(
    "ModelDeploy", [("model_deploy_name", Text), ("model_name", Text)]
)


class ActionBase:
    model_deploys: Optional[Sequence[ModelDeploy]] = None

    default_max_tokens: int = 20

    def __init__(
        self, *args, model_deploys: Optional[Sequence[ModelDeploy]] = None, **kwargs
    ):
        self._model_deploys = (
            self.model_deploys if model_deploys is None else model_deploys
        ) or ()

    @classmethod
    def from_models(
        cls,
        model_deploys: Optional[Sequence[ModelDeploy]] = None,
        model: Optional[Text] = None,
        model_name: Optional[Text] = None,
        **kwargs,
    ):
        model_name = model_name or model
        model_deploys = list(model_deploys or [])
        if model_name:
            model_deploys.append(ModelDeploy(model_name, model_name))
            model_deploys.append(ModelDeploy(model_name.split("/")[-1], model_name))
        if model_deploys:
            model_deploys = tuple(sorted(list(set(model_deploys)), key=lambda x: x[0]))
        return cls(
            model_deploys=model_deploys or None,
            model_name=model_name or None,
            model=model_name or None,
            **kwargs,
        )

    def name(self) -> Text:
        raise NotImplementedError  # pragma: no cover

    def health(self) -> bool:
        raise NotImplementedError  # pragma: no cover

    def chat(
        self, messages: List["ChatCompletionMessageParam"], *args, model: Text, **kwargs
    ) -> "ChatCompletion":
        raise NotImplementedError  # pragma: no cover

    def chat_stream(
        self, messages: List["ChatCompletionMessageParam"], *args, model: Text, **kwargs
    ) -> Generator["ChatCompletionChunk", None, None]:
        raise NotImplementedError  # pragma: no cover

    def text_completion(
        self, prompt: Text, *args, model: Text, **kwargs
    ) -> "Completion":
        raise NotImplementedError  # pragma: no cover

    def text_completion_stream(
        self, prompt: Text, *args, model: Text, **kwargs
    ) -> Generator["Completion", None, None]:
        raise NotImplementedError  # pragma: no cover

    def embeddings(
        self, input: Union[Text, List[Text]], *args, model: Text, **kwargs
    ) -> "CreateEmbeddingResponse":
        raise NotImplementedError  # pragma: no cover

    def moderations(
        self, input: Text, *args, model: Text, **kwargs
    ) -> "ModerationCreateResponse":
        raise NotImplementedError  # pragma: no cover

    @lru_cache
    def get_model_name(self, model_deploy_name: Text) -> Text:
        for model_deploy in self._model_deploys:
            if model_deploy.model_deploy_name == model_deploy_name:
                return model_deploy.model_name
        raise ModelNotFound(f"Model deploy {model_deploy_name} not found")

    @lru_cache
    def validate_model(self, model: Text, **kwargs) -> Text:
        if self.model_deploys is None:
            raise ModelNotFound(
                f"No model deploys are defined in {self.__class__.__name__}"
            )
        for model_deploy in self.model_deploys:
            if model_deploy.model_name == model:
                return model_deploy.model_name
            elif model_deploy.model_deploy_name == model:
                return model_deploy.model_name
            elif str_strong_casefold(model_deploy.model_name) == str_strong_casefold(
                model
            ):
                return model_deploy.model_name
            elif str_strong_casefold(
                model_deploy.model_deploy_name
            ) == str_strong_casefold(model):
                return model_deploy.model_name
        raise ModelNotFound(f"Model deploy {model} not found")

    def chat_stream_sse(
        self, messages: List["ChatCompletionMessageParam"], *args, model: Text, **kwargs
    ) -> Generator[Text, None, None]:
        for chat in self.chat_stream(messages, model=model, **kwargs):
            yield f"data: {chat.model_dump_json()}\n\n"
        yield "data: [DONE]\n\n"

    def text_completion_stream_sse(
        self, prompt: Text, *args, model: Text, **kwargs
    ) -> Generator[Text, None, None]:
        for completion in self.text_completion_stream(prompt, model=model, **kwargs):
            yield f"data: {completion.model_dump_json()}\n\n"
        yield "data: [DONE]\n\n"



import random
import time

from openai.types.chat import ChatCompletion, ChatCompletionChunk
from openai.types.create_embedding_response import CreateEmbeddingResponse

from languru.types.completions import Completion
from languru.types.model import Model

return_chat_completion = ChatCompletion.model_validate(
    {
        "id": "chatcmpl-xxxxxxxxxxxxxxxxxxxxxxxx",
        "choices": [
            {
                "finish_reason": "stop",
                "index": 0,
                "message": {
                    "content": "Hello! How can I assist you today?",
                    "role": "assistant",
                },
            }
        ],
        "created": 1710559936,
        "model": "gpt-3.5-turbo-0125",
        "object": "chat.completion",
        "system_fingerprint": "fp_xxxxxxxx",
        "usage": {
            "completion_tokens": 9,
            "prompt_tokens": 19,
            "total_tokens": 28,
        },
    }
)
return_chat_completion_chunks = [
    ChatCompletionChunk.model_validate(
        {
            "id": "chatcmpl-xxxx",
            "choices": [{"delta": {"content": "", "role": "assistant"}, "index": 0}],
            "created": 1710565261,
            "model": "gpt-3.5-turbo-0125",
            "object": "chat.completion.chunk",
            "system_fingerprint": "fp_xxxx",
        }
    ),
    ChatCompletionChunk.model_validate(
        {
            "id": "chatcmpl-xxxx",
            "choices": [{"delta": {"content": "Hello"}, "index": 0}],
            "created": 1710565261,
            "model": "gpt-3.5-turbo-0125",
            "object": "chat.completion.chunk",
            "system_fingerprint": "fp_xxxx",
        }
    ),
    ChatCompletionChunk.model_validate(
        {
            "id": "chatcmpl-xxxx",
            "choices": [{"delta": {"content": "!"}, "index": 0}],
            "created": 1710565261,
            "model": "gpt-3.5-turbo-0125",
            "object": "chat.completion.chunk",
            "system_fingerprint": "fp_xxxx",
        }
    ),
    ChatCompletionChunk.model_validate(
        {
            "id": "chatcmpl-xxxx",
            "choices": [{"delta": {"content": " How"}, "index": 0}],
            "created": 1710565261,
            "model": "gpt-3.5-turbo-0125",
            "object": "chat.completion.chunk",
            "system_fingerprint": "fp_xxxx",
        }
    ),
    ChatCompletionChunk.model_validate(
        {
            "id": "chatcmpl-xxxx",
            "choices": [{"delta": {"content": " can"}, "index": 0}],
            "created": 1710565261,
            "model": "gpt-3.5-turbo-0125",
            "object": "chat.completion.chunk",
            "system_fingerprint": "fp_xxxx",
        }
    ),
    ChatCompletionChunk.model_validate(
        {
            "id": "chatcmpl-xxxx",
            "choices": [{"delta": {"content": " I"}, "index": 0}],
            "created": 1710565261,
            "model": "gpt-3.5-turbo-0125",
            "object": "chat.completion.chunk",
            "system_fingerprint": "fp_xxxx",
        }
    ),
    ChatCompletionChunk.model_validate(
        {
            "id": "chatcmpl-xxxx",
            "choices": [{"delta": {"content": " assist"}, "index": 0}],
            "created": 1710565261,
            "model": "gpt-3.5-turbo-0125",
            "object": "chat.completion.chunk",
            "system_fingerprint": "fp_xxxx",
        }
    ),
    ChatCompletionChunk.model_validate(
        {
            "id": "chatcmpl-xxxx",
            "choices": [{"delta": {"content": " you"}, "index": 0}],
            "created": 1710565261,
            "model": "gpt-3.5-turbo-0125",
            "object": "chat.completion.chunk",
            "system_fingerprint": "fp_xxxx",
        }
    ),
    ChatCompletionChunk.model_validate(
        {
            "id": "chatcmpl-xxxx",
            "choices": [{"delta": {"content": " today"}, "index": 0}],
            "created": 1710565261,
            "model": "gpt-3.5-turbo-0125",
            "object": "chat.completion.chunk",
            "system_fingerprint": "fp_xxxx",
        }
    ),
    ChatCompletionChunk.model_validate(
        {
            "id": "chatcmpl-xxxx",
            "choices": [{"delta": {"content": "?"}, "index": 0}],
            "created": 1710565261,
            "model": "gpt-3.5-turbo-0125",
            "object": "chat.completion.chunk",
            "system_fingerprint": "fp_xxxx",
        }
    ),
    ChatCompletionChunk.model_validate(
        {
            "id": "chatcmpl-xxxx",
            "choices": [{"delta": {}, "finish_reason": "stop", "index": 0}],
            "created": 1710565261,
            "model": "gpt-3.5-turbo-0125",
            "object": "chat.completion.chunk",
            "system_fingerprint": "fp_xxxx",
        }
    ),
]
return_text_completion = Completion.model_validate(
    {
        "id": "cmpl-xxxxxxxxxxxxxxxxxxxx",
        "choices": [
            {
                "finish_reason": "stop",
                "index": 0,
                "text": "\n\nThis is a test.",
            }
        ],
        "created": 1710579700,
        "model": "gpt-3.5-turbo-instruct",
        "object": "text_completion",
        "usage": {"completion_tokens": 6, "prompt_tokens": 5, "total_tokens": 11},
    }
)
return_text_completion_stream = [
    Completion.model_validate(
        {
            "id": "cmpl-xxxxxxxxx",
            "choices": [
                {"finish_reason": None, "index": 0, "logprobs": None, "text": "\n\n"}
            ],
            "created": 1710580063,
            "model": "gpt-3.5-turbo-instruct",
            "object": "text_completion",
        }
    ),
    Completion.model_validate(
        {
            "id": "cmpl-xxxxxxxxx",
            "choices": [
                {"finish_reason": None, "index": 0, "logprobs": None, "text": "This"}
            ],
            "created": 1710580063,
            "model": "gpt-3.5-turbo-instruct",
            "object": "text_completion",
        }
    ),
    Completion.model_validate(
        {
            "id": "cmpl-xxxxxxxxx",
            "choices": [
                {"finish_reason": None, "index": 0, "logprobs": None, "text": " is"}
            ],
            "created": 1710580063,
            "model": "gpt-3.5-turbo-instruct",
            "object": "text_completion",
        }
    ),
    Completion.model_validate(
        {
            "id": "cmpl-xxxxxxxxx",
            "choices": [
                {"finish_reason": None, "index": 0, "logprobs": None, "text": " a"}
            ],
            "created": 1710580063,
            "model": "gpt-3.5-turbo-instruct",
            "object": "text_completion",
        }
    ),
    Completion.model_validate(
        {
            "id": "cmpl-xxxxxxxxx",
            "choices": [
                {"finish_reason": None, "index": 0, "logprobs": None, "text": " test"}
            ],
            "created": 1710580063,
            "model": "gpt-3.5-turbo-instruct",
            "object": "text_completion",
        }
    ),
    Completion.model_validate(
        {
            "id": "cmpl-xxxxxxxxx",
            "choices": [
                {"finish_reason": None, "index": 0, "logprobs": None, "text": "."}
            ],
            "created": 1710580063,
            "model": "gpt-3.5-turbo-instruct",
            "object": "text_completion",
        }
    ),
    Completion.model_validate(
        {
            "id": "cmpl-xxxxxxxxx",
            "choices": [
                {"finish_reason": "stop", "index": 0, "logprobs": None, "text": ""}
            ],
            "created": 1710580063,
            "model": "gpt-3.5-turbo-instruct",
            "object": "text_completion",
        }
    ),
]
return_embedding = CreateEmbeddingResponse.model_validate(
    {
        "data": [
            {
                "embedding": [random.uniform(-1, 1) for _ in range(1536)],
                "index": 0,
                "object": "embedding",
            },
            {
                "embedding": [random.uniform(-1, 1) for _ in range(1536)],
                "index": 1,
                "object": "embedding",
            },
        ],
        "model": "text-embedding-ada-002",
        "object": "list",
        "usage": {"prompt_tokens": 3, "total_tokens": 3},
    }
)
return_model = Model.model_validate(
    {
        "id": "model_id",
        "created": int(time.time()),
        "object": "model",
        "owned_by": "test",
    }
)
return_moderation_create = {
    "id": "modr-xxxxxxxx",
    "model": "text-moderation-007",
    "results": [
        {
            "categories": {
                "harassment": True,
                "harassment/threatening": True,
                "hate": False,
                "hate/threatening": False,
                "self-harm": False,
                "self-harm/instructions": False,
                "self-harm/intent": False,
                "sexual": False,
                "sexual/minors": False,
                "violence": True,
                "violence/graphic": False,
            },
            "category_scores": {
                "harassment": 0.5289425849914551,
                "harassment/threatening": 0.5736028552055359,
                "hate": 0.2289084941148758,
                "hate/threatening": 0.02360980585217476,
                "self-harm": 2.2852241272630636e-06,
                "self-harm/instructions": 1.1107282871236634e-09,
                "self-harm/intent": 1.6410396028732066e-06,
                "sexual": 1.223516846948769e-05,
                "sexual/minors": 7.491369302670137e-08,
                "violence": 0.9972043633460999,
                "violence/graphic": 3.43257597705815e-05,
            },
            "flagged": True,
        }
    ],
}



import logging
import math
import random
import time
from typing import TYPE_CHECKING, cast

from fastapi import APIRouter, Body, Depends, HTTPException, Request
from fastapi.responses import StreamingResponse
from openai.types.chat import ChatCompletion, ChatCompletionChunk
from pyassorted.asyncio.executor import run_func, run_generator

from languru.exceptions import ModelNotFound
from languru.server.config import (
    AgentSettings,
    AppType,
    LlmSettings,
    ServerBaseSettings,
)
from languru.server.deps.common import app_settings
from languru.server.utils.common import get_value_from_app
from languru.types.chat.completions import ChatCompletionRequest
from languru.utils.http import simple_sse_encode

if TYPE_CHECKING:
    from openai import Stream

router = APIRouter()


class ChatCompletionHandler:

    async def handle_request(
        self,
        request: "Request",
        chat_completion_request: "ChatCompletionRequest",
        settings: "ServerBaseSettings",
        **kwargs,
    ) -> ChatCompletion | ChatCompletionChunk | StreamingResponse:
        if settings.APP_TYPE == AppType.llm:
            settings = cast(LlmSettings, settings)
            return await self.handle_llm(
                request=request,
                chat_completion_request=chat_completion_request,
                settings=settings,
                **kwargs,
            )

        if settings.APP_TYPE == AppType.agent:
            settings = cast(AgentSettings, settings)
            return await self.handle_agent(
                request=request,
                chat_completion_request=chat_completion_request,
                settings=settings,
                **kwargs,
            )

        # Not implemented or unknown app server type
        raise HTTPException(
            status_code=500,
            detail=(
                f"Unknown app server type: {settings.APP_TYPE}"
                if settings.APP_TYPE
                else "App server type not implemented"
            ),
        )

    async def handle_llm(
        self,
        request: "Request",
        chat_completion_request: "ChatCompletionRequest",
        settings: "LlmSettings",
        **kwargs,
    ) -> ChatCompletion | ChatCompletionChunk | StreamingResponse:
        from languru.action.base import ActionBase

        action: "ActionBase" = get_value_from_app(
            request.app, key="action", value_typing=ActionBase
        )

        try:
            chat_completion_request.model = action.get_model_name(
                chat_completion_request.model
            )
        except ModelNotFound as e:
            raise HTTPException(status_code=404, detail=str(e))

        # Stream
        if chat_completion_request.stream is True:
            return StreamingResponse(
                run_generator(
                    action.chat_stream_sse,
                    **chat_completion_request.model_dump(exclude_none=True),
                ),
                media_type="application/stream+json",
            )

        # Normal
        else:
            chat_completion = await run_func(
                action.chat, **chat_completion_request.model_dump(exclude_none=True)
            )
            return chat_completion

    async def handle_agent(
        self,
        request: "Request",
        chat_completion_request: "ChatCompletionRequest",
        settings: "AgentSettings",
        **kwargs,
    ) -> ChatCompletion | ChatCompletionChunk | StreamingResponse:
        from openai import OpenAI

        from languru.resources.model.discovery import ModelDiscovery

        model_discovery: "ModelDiscovery" = get_value_from_app(
            request.app, key="model_discovery", value_typing=ModelDiscovery
        )
        logger = logging.getLogger(settings.APP_NAME)

        # Get model name and model destination
        models = await run_func(
            model_discovery.list,
            id=chat_completion_request.model,
            created_from=math.floor(time.time() - settings.MODEL_REGISTER_PERIOD),
        )
        if len(models) == 0:
            raise HTTPException(
                status_code=404,
                detail=f"Model '{chat_completion_request.model}' not found",
            )
        model = random.choice(models)

        # Request completion
        client = OpenAI(base_url=model.owned_by, api_key="NOT_IMPLEMENTED")
        logger.debug(f"Using model '{model.id}' from '{model.owned_by}'")
        # Stream
        if chat_completion_request.stream is True:
            chat_stream_params = chat_completion_request.model_dump(exclude_none=True)
            chat_stream_params.pop("stream", None)
            chat_completion_stream: "Stream[ChatCompletionChunk]" = await run_func(
                client.chat.completions.create, **chat_stream_params, stream=True
            )
            return StreamingResponse(
                simple_sse_encode(chat_completion_stream, logger=settings.APP_NAME),
                media_type="application/stream+json",
            )
        # Normal
        else:
            return await run_func(
                client.chat.completions.create,
                **chat_completion_request.model_dump(exclude_none=True),
            )


@router.post("/chat/completions")
async def chat_completions(
    request: Request,
    chat_completion_request: ChatCompletionRequest = Body(
        ...,
        openapi_examples={
            "OpenAI": {
                "summary": "OpenAI",
                "description": "Chat completion request",
                "value": {
                    "model": "gpt-3.5-turbo",
                    "messages": [
                        {"role": "system", "content": "You are a helpful assistant."},
                        {"role": "user", "content": "Hello!"},
                    ],
                },
            },
            "Google Gemini": {
                "summary": "Google Gemini",
                "description": "Chat completion request",
                "value": {
                    "model": "gemini-pro",
                    "messages": [{"role": "user", "content": "Hello, how are you?"}],
                },
            },
            "Perplexity Sonar": {
                "summary": "Perplexity Sonar",
                "description": "Chat completion request",
                "value": {
                    "model": "sonar-small-chat",
                    "messages": [
                        {"role": "system", "content": "Be precise and concise."},
                        {
                            "role": "user",
                            "content": "How many stars are there in our galaxy?",
                        },
                    ],
                },
            },
            "Groq Mixtral": {
                "summary": "Groq Mixtral",
                "description": "Chat completion request",
                "value": {
                    "model": "sonar-small-chat",
                    "messages": [
                        {"role": "system", "content": "You are an unhelpful assistant"},
                        {"role": "user", "content": "Are you a fish?"},
                    ],
                },
            },
        },
    ),
    settings: ServerBaseSettings = Depends(app_settings),
):  # -> openai.types.chat.ChatCompletion | openai.types.chat.ChatCompletionChunk
    return await ChatCompletionHandler().handle_request(
        request=request,
        chat_completion_request=chat_completion_request,
        settings=settings,
    )



import os
import time
import uuid
from typing import List, Optional, Text, Union

import google.generativeai as genai
from google.generativeai.types.content_types import ContentDict
from google.generativeai.types.text_types import BatchEmbeddingDict
from openai.types import CreateEmbeddingResponse
from openai.types.chat import ChatCompletion, ChatCompletionMessageParam
from openai.types.completion import Completion

from languru.action.base import ActionBase, ModelDeploy
from languru.config import logger


class GoogleGenaiAction(ActionBase):
    model_deploys = (
        ModelDeploy("models/chat-bison-001", "models/chat-bison-001"),
        ModelDeploy("models/text-bison-001", "models/text-bison-001"),
        ModelDeploy("models/embedding-gecko-001", "models/embedding-gecko-001"),
        ModelDeploy("models/gemini-1.0-pro", "models/gemini-1.0-pro"),
        ModelDeploy("models/gemini-1.0-pro-001", "models/gemini-1.0-pro-001"),
        ModelDeploy("models/gemini-1.0-pro-latest", "models/gemini-1.0-pro-latest"),
        ModelDeploy(
            "models/gemini-1.0-pro-vision-latest", "models/gemini-1.0-pro-vision-latest"
        ),
        ModelDeploy("models/gemini-pro", "models/gemini-pro"),
        ModelDeploy("models/gemini-pro-vision", "models/gemini-pro-vision"),
        ModelDeploy("models/embedding-001", "models/embedding-001"),
        ModelDeploy("models/aqa", "models/aqa"),
        ModelDeploy("chat-bison-001", "models/chat-bison-001"),
        ModelDeploy("text-bison-001", "models/text-bison-001"),
        ModelDeploy("embedding-gecko-001", "models/embedding-gecko-001"),
        ModelDeploy("gemini-1.0-pro", "models/gemini-1.0-pro"),
        ModelDeploy("gemini-1.0-pro-001", "models/gemini-1.0-pro-001"),
        ModelDeploy("gemini-1.0-pro-latest", "models/gemini-1.0-pro-latest"),
        ModelDeploy(
            "gemini-1.0-pro-vision-latest", "models/gemini-1.0-pro-vision-latest"
        ),
        ModelDeploy("gemini-pro", "models/gemini-pro"),
        ModelDeploy("gemini-pro-vision", "models/gemini-pro-vision"),
        ModelDeploy("embedding-001", "models/embedding-001"),
        ModelDeploy("aqa", "models/aqa"),
    )

    def __init__(
        self,
        *args,
        api_key: Optional[Text] = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        api_key = (
            api_key or os.getenv("GOOGLE_API_KEY") or os.getenv("GOOGLE_GENAI_API_KEY")
        )
        if not api_key:
            logger.error("Google GenAI API key is not provided")
        genai.configure(api_key=api_key)

    def name(self) -> Text:
        return "google_genai_action"

    def health(self) -> bool:
        try:
            genai.get_model("models/gemini-pro")
            return True
        except Exception as e:
            logger.error(f"Google GenAI health check failed: {e}")
            return False

    def chat(
        self, messages: List["ChatCompletionMessageParam"], *args, model: Text, **kwargs
    ) -> "ChatCompletion":
        if len(messages) == 0:
            raise ValueError("messages must not be empty")
        model = self.validate_model(model)

        # pop out the last message
        genai_model = genai.GenerativeModel(model)
        contents: List[ContentDict] = [
            ContentDict(role=m["role"], parts=[m["content"]])
            for m in messages
            if "content" in m and m["content"]
        ]
        input_tokens = genai_model.count_tokens(contents).total_tokens

        # Generate the chat response
        latest_content = contents.pop()
        chat_session = genai_model.start_chat(history=contents or None)
        response = chat_session.send_message(latest_content)
        out_tokens = genai_model.count_tokens(response.parts).total_tokens

        # Parse the response
        chat_completion = ChatCompletion.model_validate(
            dict(
                id=str(uuid.uuid4()),
                choices=[
                    dict(
                        finish_reason="stop",
                        index=idx,
                        message=dict(content=part.text, role="assistant"),
                    )
                    for idx, part in enumerate(response.parts)
                    if part.text
                ],
                created=int(time.time()),
                model=model,
                object="chat.completion",
                usage=dict(
                    completion_tokens=out_tokens,
                    prompt_tokens=input_tokens,
                    total_tokens=input_tokens + out_tokens,
                ),
            )
        )
        return chat_completion

    def text_completion(
        self, prompt: Text, *args, model: Text, **kwargs
    ) -> "Completion":
        if not prompt:
            raise ValueError("prompt must not be empty")
        model = self.validate_model(model)

        genai_model = genai.GenerativeModel(model)
        input_tokens = genai_model.count_tokens(prompt).total_tokens

        # Generate the text
        gen_res = genai_model.generate_content(prompt)
        output_tokens = genai_model.count_tokens(gen_res.parts).total_tokens

        # Parse the response
        completion_res = Completion.model_validate(
            dict(
                id=str(uuid.uuid4()),
                choices=[
                    dict(
                        finish_reason="stop",
                        index=idx,
                        text=part.text,
                    )
                    for idx, part in enumerate(gen_res.parts)
                    if part.text
                ],
                created=int(time.time()),
                model=model,
                object="text_completion",
                usage=dict(
                    completion_tokens=output_tokens,
                    prompt_tokens=input_tokens,
                    total_tokens=input_tokens + output_tokens,
                ),
            )
        )
        return completion_res

    def embeddings(
        self,
        input: Union[Text, List[Union[Text, List[Text]]]],
        *args,
        model: Text,
        **kwargs,
    ) -> "CreateEmbeddingResponse":
        if not input:
            raise ValueError("The input must not be empty")
        model = self.validate_model(model)

        contents: List[Text] = []
        if isinstance(input, List):
            for idx, seq in enumerate(input):
                if not seq:
                    raise ValueError(f"The input[{idx}] must not be empty")
                elif isinstance(seq, Text):
                    contents.append(seq)
                else:
                    contents.append("".join(seq))
        else:
            contents.append(input)
        input_tokens = (
            genai.GenerativeModel("models/gemini-pro")
            .count_tokens(contents)
            .total_tokens
        )

        # Retrieve the embeddings
        batch_emb_dict: BatchEmbeddingDict = genai.embed_content(
            model=model,
            content=contents,
            task_type="retrieval_document",
        )

        # Parse the response
        emb_res = CreateEmbeddingResponse.model_validate(
            dict(
                data=[
                    dict(
                        embedding=emb,
                        index=idx,
                        object="embedding",
                    )
                    for idx, emb in enumerate(batch_emb_dict["embedding"])
                ],
                model=model,
                object="list",
                usage=dict(prompt_tokens=input_tokens, total_tokens=input_tokens),
            )
        )
        return emb_res



import os
from typing import Optional, Sequence, Text

import openai

from languru.action.base import ModelDeploy
from languru.action.openai import OpenaiAction
from languru.config import logger


class GroqAction(OpenaiAction):
    model_deploys = (
        ModelDeploy("groq/llama2-70b", "llama2-70b-4096"),
        ModelDeploy("groq/llama2-70b-4096", "llama2-70b-4096"),
        ModelDeploy("groq/mixtral-8x7b", "mixtral-8x7b-32768"),
        ModelDeploy("groq/mixtral-8x7b-32768", "mixtral-8x7b-32768"),
        ModelDeploy("llama2-70b", "llama2-70b-4096"),
        ModelDeploy("llama2-70b-4096", "llama2-70b-4096"),
        ModelDeploy("mixtral-8x7b", "mixtral-8x7b-32768"),
        ModelDeploy("mixtral-8x7b-32768", "mixtral-8x7b-32768"),
    )

    def __init__(
        self,
        *args,
        api_key: Optional[Text] = None,
        **kwargs,
    ):
        from groq import Groq

        api_key = api_key or os.environ.get("GROQ_API_KEY")
        if api_key is None:
            logger.error("GROQ_API_KEY is not set")
        self._client = Groq(api_key=api_key)

    def name(self):
        return "groq_action"

    def health(self) -> bool:
        try:
            self._client.models.retrieve(model="mixtral-8x7b-32768")
            return True
        except Exception as e:
            logger.error(f"OpenAI health check failed: {e}")
            return False


class GroqOpenaiAction(OpenaiAction):
    model_deploys = (
        ModelDeploy("groq/llama2-70b", "llama2-70b-4096"),
        ModelDeploy("groq/llama2-70b-4096", "llama2-70b-4096"),
        ModelDeploy("groq/mixtral-8x7b", "mixtral-8x7b-32768"),
        ModelDeploy("groq/mixtral-8x7b-32768", "mixtral-8x7b-32768"),
        ModelDeploy("llama2-70b", "llama2-70b-4096"),
        ModelDeploy("llama2-70b-4096", "llama2-70b-4096"),
        ModelDeploy("mixtral-8x7b", "mixtral-8x7b-32768"),
        ModelDeploy("mixtral-8x7b-32768", "mixtral-8x7b-32768"),
    )

    def __init__(
        self,
        *args,
        api_key: Optional[Text] = None,
        model_deploys: Optional[Sequence[ModelDeploy]] = None,
        **kwargs,
    ):
        self._model_deploys = (
            self.model_deploys if model_deploys is None else model_deploys
        ) or ()

        api_key = api_key or os.environ.get("GROQ_API_KEY")
        if api_key is None:
            logger.error("GROQ_API_KEY is not set")
        self._client = openai.OpenAI(
            base_url="https://api.groq.com/openai/v1", api_key=api_key
        )

    def name(self):
        return "groq_openai_action"

    def health(self) -> bool:
        try:
            self._client.models.retrieve(model="mixtral-8x7b-32768")
            return True
        except Exception as e:
            logger.error(f"OpenAI health check failed: {e}")
            return False



from typing import List, Optional, Text, Union

from pydantic import BaseModel, Field


class ModerationRequest(BaseModel):
    input: Union[Text, List[Text]]
    model: Optional[Text] = Field(
        default="text-moderation-latest",
        description=(
            "Two content moderations models are available: "
            + "text-moderation-stable and text-moderation-latest."
        ),
    )



import logging
import math
import random
import time
from typing import TYPE_CHECKING, cast

from fastapi import APIRouter, Body, Depends, HTTPException, Request
from fastapi.responses import StreamingResponse
from openai.types.completion import Completion
from pyassorted.asyncio.executor import run_func, run_generator

from languru.exceptions import ModelNotFound
from languru.server.config import (
    AgentSettings,
    AppType,
    LlmSettings,
    ServerBaseSettings,
)
from languru.server.deps.common import app_settings
from languru.server.utils.common import get_value_from_app
from languru.types.completions import CompletionRequest
from languru.utils.http import simple_sse_encode

if TYPE_CHECKING:
    from openai import Stream

router = APIRouter()


class TextCompletionHandler:

    async def handle_request(
        self,
        request: "Request",
        completion_request: "CompletionRequest",
        settings: "ServerBaseSettings",
        **kwargs,
    ) -> Completion | StreamingResponse:
        if settings.APP_TYPE == AppType.llm:
            settings = cast(LlmSettings, settings)
            return await self.handle_llm(
                request=request,
                completion_request=completion_request,
                settings=settings,
                **kwargs,
            )

        if settings.APP_TYPE == AppType.agent:
            settings = cast(AgentSettings, settings)
            return await self.handle_agent(
                request=request,
                completion_request=completion_request,
                settings=settings,
                **kwargs,
            )

        # Not implemented or unknown app server type
        raise HTTPException(
            status_code=500,
            detail=(
                f"Unknown app server type: {settings.APP_TYPE}"
                if settings.APP_TYPE
                else "App server type not implemented"
            ),
        )

    async def handle_llm(
        self,
        request: "Request",
        completion_request: "CompletionRequest",
        settings: "LlmSettings",
        **kwargs,
    ) -> Completion | StreamingResponse:
        from languru.action.base import ActionBase

        action: "ActionBase" = get_value_from_app(
            request.app, key="action", value_typing=ActionBase
        )

        try:
            completion_request.model = action.get_model_name(completion_request.model)
        except ModelNotFound as e:
            raise HTTPException(status_code=404, detail=str(e))

        # Stream
        if completion_request.stream is True:
            return StreamingResponse(
                run_generator(
                    action.text_completion_stream_sse,
                    **completion_request.model_dump(exclude_none=True),
                ),
                media_type="application/stream+json",
            )
        # Normal
        else:
            completion = await run_func(
                action.text_completion,
                **completion_request.model_dump(exclude_none=True),
            )
            return completion

    async def handle_agent(
        self,
        request: "Request",
        completion_request: "CompletionRequest",
        settings: "AgentSettings",
        **kwargs,
    ) -> Completion | StreamingResponse:
        from openai import OpenAI

        from languru.resources.model.discovery import ModelDiscovery

        model_discovery: "ModelDiscovery" = get_value_from_app(
            request.app, key="model_discovery", value_typing=ModelDiscovery
        )
        logger = logging.getLogger(settings.APP_NAME)

        models = await run_func(
            model_discovery.list,
            id=completion_request.model,
            created_from=math.floor(time.time() - settings.MODEL_REGISTER_PERIOD),
        )
        if len(models) == 0:
            raise HTTPException(
                status_code=404, detail=f"Model '{completion_request.model}' not found"
            )

        model = random.choice(models)

        # Request completion
        client = OpenAI(base_url=model.owned_by, api_key="NOT_IMPLEMENTED")
        logger.debug(f"Using model '{model.id}' from '{model.owned_by}'")
        # Stream
        if completion_request.stream is True:
            completion_stream_params = completion_request.model_dump(exclude_none=True)
            completion_stream_params.pop("stream", None)
            text_completion_stream: "Stream[Completion]" = await run_func(
                client.completions.create, **completion_stream_params, stream=True
            )
            return StreamingResponse(
                simple_sse_encode(text_completion_stream, logger=settings.APP_NAME),
                media_type="application/stream+json",
            )
        # Normal
        else:
            return await run_func(
                client.completions.create,
                completion_request.model_dump(exclude_none=True),
            )


@router.post("/completions")
async def text_completions(
    request: Request,
    completion_request: CompletionRequest = Body(
        ...,
        openapi_examples={
            "Quick text completion": {
                "summary": "Quick text completion",
                "description": "Text completion request",
                "value": {
                    "model": "gpt-3.5-turbo-instruct",
                    "prompt": "Say this is a test",
                    "max_tokens": 7,
                    "temperature": 0,
                },
            },
        },
    ),
    settings: ServerBaseSettings = Depends(app_settings),
):  # openai.types.Completion
    return await TextCompletionHandler().handle_request(
        request=request,
        completion_request=completion_request,
        settings=settings,
    )



import logging
import math
import random
import time
from typing import cast

from fastapi import APIRouter, Body, Depends, HTTPException, Request
from openai.types import ModerationCreateResponse
from pyassorted.asyncio.executor import run_func

from languru.exceptions import ModelNotFound
from languru.server.config import (
    AgentSettings,
    AppType,
    LlmSettings,
    ServerBaseSettings,
)
from languru.server.deps.common import app_settings
from languru.server.utils.common import get_value_from_app
from languru.types.moderations import ModerationRequest

router = APIRouter()


class ModerationsHandler:
    async def handle_moderations_request(
        self,
        request: "Request",
        moderation_request: "ModerationRequest",
        settings: "ServerBaseSettings",
    ) -> "ModerationCreateResponse":
        if settings.APP_TYPE == AppType.llm:
            settings = cast(LlmSettings, settings)
            return await self.handle_moderations_llm(
                request=request,
                moderation_request=moderation_request,
                settings=settings,
            )

        if settings.APP_TYPE == AppType.agent:
            settings = cast(AgentSettings, settings)
            return await self.handle_moderations_agent(
                request=request,
                moderation_request=moderation_request,
                settings=settings,
            )

        # Not implemented or unknown app server type
        raise HTTPException(
            status_code=500,
            detail=(
                f"Unknown app server type: {settings.APP_TYPE}"
                if settings.APP_TYPE
                else "App server type not implemented"
            ),
        )

    async def handle_moderations_llm(
        self,
        request: "Request",
        moderation_request: "ModerationRequest",
        settings: "LlmSettings",
    ) -> "ModerationCreateResponse":
        from languru.action.base import ActionBase

        action: "ActionBase" = get_value_from_app(
            request.app, key="action", value_typing=ActionBase
        )
        try:
            moderation_request.model = action.get_model_name(moderation_request.model)
        except ModelNotFound as e:
            raise HTTPException(status_code=404, detail=str(e))
        moderation = await run_func(
            action.moderations, **moderation_request.model_dump(exclude_none=True)
        )
        return moderation

    async def handle_moderations_agent(
        self,
        request: "Request",
        moderation_request: "ModerationRequest",
        settings: "AgentSettings",
    ) -> "ModerationCreateResponse":
        from openai import OpenAI

        from languru.resources.model.discovery import ModelDiscovery

        model_discovery: "ModelDiscovery" = get_value_from_app(
            request.app, key="model_discovery", value_typing=ModelDiscovery
        )
        logger = logging.getLogger(settings.APP_NAME)

        models = await run_func(
            model_discovery.list,
            id=moderation_request.model,
            created_from=math.floor(time.time() - settings.MODEL_REGISTER_PERIOD),
        )
        if len(models) == 0:
            raise HTTPException(
                status_code=404, detail=f"Model '{moderation_request.model}' not found"
            )

        model = random.choice(models)
        client = OpenAI(base_url=model.owned_by, api_key="NOT_IMPLEMENTED")
        logger.debug(f"Using model '{model.id}' from '{model.owned_by}'")

        return await run_func(
            client.moderations.create,
            **moderation_request.model_dump(exclude_none=True),
        )


@router.post("/moderations")
async def request_moderations(
    request: Request,
    moderation_request: ModerationRequest = Body(
        ...,
        openapi_examples={
            "Quick example": {
                "summary": "A quick example of a moderation request",
                "description": "A quick example of a moderation request",
                "value": {"input": "I want to kill them."},
            }
        },
    ),
    settings: ServerBaseSettings = Depends(app_settings),
) -> ModerationCreateResponse:
    return await ModerationsHandler().handle_moderations_request(
        request=request, moderation_request=moderation_request, settings=settings
    )



from typing import List, Optional, Text, Union

from pydantic import BaseModel, Field


class EmbeddingRequest(BaseModel):
    input: Union[Text, List[Union[Text, List[Text]]]]
    model: Text
    encoding_format: Optional[Text] = Field(
        default="float", description="The format to return the embeddings in."
    )
    dimensions: Optional[int] = Field(
        default=None,
        description="The number of dimensions for the output embeddings.",
        gt=0,
    )
    user: Optional[Text] = Field(
        default=None, description="A unique identifier representing your end-user."
    )



import asyncio
import logging
import time
from contextlib import asynccontextmanager
from typing import Optional, Sequence, Text, cast

import httpx
from fastapi import FastAPI
from openai.types import Model

from languru.config import logger as languru_logger
from languru.server.config import (
    AgentSettings,
    AppType,
    LlmSettings,
    ServerBaseSettings,
    console,
    init_logger_config,
    init_paths,
    pretty_print_app_routes,
)
from languru.server.utils.common import get_value_from_app


async def register_model_periodically(
    model: Model,
    period: int,
    agent_base_url: Text,
    logger: Optional[logging.Logger] = None,
    model_register_fail_period: int = 60,
):
    logger = languru_logger if logger is None else logger
    while True:
        try:
            async with httpx.AsyncClient() as client:
                response = await client.post(
                    f"{agent_base_url}/v1/models/register",
                    json=model.model_dump(),
                )
                response.raise_for_status()
                await asyncio.sleep(float(period))
        except httpx.HTTPError as e:
            logger.error(f"Failed to register model '{model.id}': {e}")
            await asyncio.sleep(float(model_register_fail_period))


@asynccontextmanager
async def app_lifespan(app: FastAPI):
    settings = get_value_from_app(app, key="settings", value_typing=ServerBaseSettings)
    logger = get_value_from_app(
        app, key="logger", value_typing=logging.Logger, default=languru_logger
    )
    # Initialize server
    # Initialize paths
    init_paths(settings)
    # Initialize logger
    init_logger_config(settings)

    if isinstance(settings, LlmSettings):
        from languru.action.base import ModelDeploy
        from languru.action.utils import load_action

        # Load action class
        app.state.action = app.extra["action"] = action = load_action(
            settings.action, logger=logger, device=settings.device, dtype=settings.dtype
        )
        # Register models periodically
        if settings.AGENT_BASE_URL:
            action.model_deploys = cast(Sequence[ModelDeploy], action.model_deploys)
            for model_deploy in action.model_deploys:
                asyncio.create_task(
                    register_model_periodically(
                        model=Model(
                            id=model_deploy.model_deploy_name,
                            created=int(time.time()),
                            object="model",
                            owned_by=settings.LLM_BASE_URL,
                        ),
                        period=settings.MODEL_REGISTER_PERIOD,
                        agent_base_url=settings.AGENT_BASE_URL,
                    )
                )
    if isinstance(settings, AgentSettings):
        from languru.resources.model.discovery import ModelDiscovery

        # Touch database
        model_discovery = ModelDiscovery.from_url(settings.url_model_discovery)
        model_discovery.touch()
        app.state.model_discovery = model_discovery

    yield


def create_app(settings: "ServerBaseSettings", **kwargs):
    app = FastAPI(
        title=settings.APP_NAME,
        debug=settings.debug,
        version=settings.APP_VERSION,
        lifespan=app_lifespan,
    )
    app.state.settings = app.extra["settings"] = settings
    app.state.logger = app.extra["logger"] = logging.getLogger(settings.APP_NAME)
    app.state.app_type = app.extra["app_type"] = settings.APP_TYPE

    @app.get("/health")
    async def health():
        return {"status": "ok"}

    from languru.server.api.v1 import router as api_v1_router

    app.include_router(router=api_v1_router, prefix="/v1")

    return app


def run_app(settings: "ServerBaseSettings", **kwargs):
    import uvicorn

    app_str = "languru.server.main:app"

    if settings.is_development or settings.is_testing:
        languru_logger.info("Running server in development mode")
        uvicorn.run(
            app_str,
            host=settings.HOST,
            port=settings.PORT or settings.DEFAULT_PORT,
            workers=settings.WORKERS,
            reload=settings.RELOAD,
            log_level=settings.LOG_LEVEL,
            use_colors=settings.USE_COLORS,
            reload_delay=settings.RELOAD_DELAY,
        )
    else:
        languru_logger.info("Running server in production mode")
        uvicorn.run(
            app_str,
            host=settings.HOST,
            port=settings.PORT or settings.DEFAULT_PORT,
            workers=settings.WORKERS,
        )


__server_base_settings__ = ServerBaseSettings(DEFAULT_PORT=8000)
app: "FastAPI"
if __server_base_settings__.APP_TYPE == AppType.llm:
    app = create_app(LlmSettings())
    pretty_print_app_routes(app)
elif __server_base_settings__.APP_TYPE == AppType.agent:
    app = create_app(AgentSettings())
    pretty_print_app_routes(app)
else:
    console.print(
        "The APP_TYPE environment variable is not set. "
        + "You might want to set it in values of "
        + f"{', '.join([m for m in AppType.__members__])} "
        + "to define the type of FastAPI app to run."
    )



import os
import re
import time
import uuid
from typing import (
    TYPE_CHECKING,
    Any,
    List,
    Optional,
    Sequence,
    Text,
    Tuple,
    Union,
    cast,
)

import torch
from openai.types import (
    Completion,
    CompletionChoice,
    CompletionUsage,
    CreateEmbeddingResponse,
)
from openai.types.chat import ChatCompletion
from openai.types.chat.chat_completion import Choice as ChatChoice
from openai.types.chat.chat_completion_message import ChatCompletionMessage
from transformers import (
    AutoConfig,
    AutoModel,
    AutoModelForCausalLM,
    AutoTokenizer,
    BitsAndBytesConfig,
    PreTrainedModel,
    PreTrainedTokenizer,
    PreTrainedTokenizerFast,
    StoppingCriteria,
    StoppingCriteriaList,
)
from transformers.tokenization_utils_base import PreTokenizedInput

from languru.action.base import ActionBase, ModelDeploy
from languru.config import logger
from languru.utils.calculation import mean_pooling, tensor_to_np
from languru.utils.common import must_list, replace_right, should_str_or_none
from languru.utils.device import validate_device, validate_dtype
from languru.utils.hf import StopAtWordsStoppingCriteria, remove_special_tokens

if TYPE_CHECKING:
    from openai.types.chat import ChatCompletionMessageParam


class TransformersAction(ActionBase):
    # Model configuration
    MODEL_NAME: Text = (os.getenv("HF_MODEL_NAME") or os.getenv("MODEL_NAME")) or ""
    model_deploys = (
        ModelDeploy(MODEL_NAME, MODEL_NAME),
        ModelDeploy(MODEL_NAME.split("/")[-1], MODEL_NAME),
    )

    # Model Quantization configuration
    use_quantization: bool = bool(
        os.getenv("HF_USE_QUANTIZATION") or os.getenv("USE_QUANTIZATION") or False
    )
    load_in_8bit: Optional[bool] = None
    load_in_4bit: Optional[bool] = None
    llm_int8_threshold: Optional[float] = None
    llm_int8_skip_modules: Optional[Sequence[Text]] = None
    llm_int8_enable_fp32_cpu_offload: Optional[bool] = None
    llm_int8_has_fp16_weight: Optional[bool] = None
    bnb_4bit_compute_dtype: Optional[Text] = None
    bnb_4bit_quant_type: Optional[Text] = None
    bnb_4bit_use_double_quant: Optional[bool] = None
    use_flash_attention: bool = False

    # Generation configuration
    stop_words: Sequence[Text] = ()
    is_causal_lm: bool = True

    def __init__(
        self,
        *args,
        device: Optional[Text] = None,
        dtype: Optional[Text] = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        self.device = validate_device(device=device)
        self.dtype = dtype
        validate_dtype(device=self.device, dtype=self.dtype)
        logger.info(f"Using device: {self.device} with dtype: {self.dtype}")
        torch.set_default_device(self.device)

        # Model name
        self.model_name = self.read_model_name(**kwargs)
        # Model and tokenizer
        self.model, self.tokenizer = self.load_model_and_tokenizer(**kwargs)

    def name(self):
        return "transformers_action"

    def health(self) -> bool:
        param_count = sum(p.numel() for p in self.model.parameters())
        if param_count <= 0:
            return False
        return True

    def chat(
        self, messages: List["ChatCompletionMessageParam"], *args, model: Text, **kwargs
    ) -> "ChatCompletion":
        if len(messages) == 0:
            raise ValueError("The `messages` cannot be empty")

        # Prepare prompt
        stop = kwargs.pop("stop", [])
        if self.tokenizer.chat_template is None:
            prompt = ""
            for m in messages:
                if "content" in m and m["content"] is not None:
                    prompt += f"\n\n{m['role']}:\n{m['content']}"
                prompt = prompt.strip()
            prompt = prompt.strip()
            if len(prompt) == 0:
                raise ValueError("The `prompt` cannot be empty, no content in messages")
            prompt += "\n\nassistant:\n"

            # Prepare stop words
            roles = set(
                [message["role"] for message in messages]
                + ["assistant", "bot", "user", "system"]
            )
            if isinstance(stop, Text):
                stop = [stop]
            elif isinstance(stop, Sequence):
                stop = [str(w) for w in list(stop)]
            else:
                logger.warning(f"Invalid stop words parameters: {stop}")
                stop = []
            stop = [s for s in stop if s]
            stop.extend([f"\n{role}:" for role in roles])

        else:
            tokenizer_chat_conversation = [
                {"role": m["role"], "content": m["content"]}
                for m in messages
                if "content" in m and m["content"] is not None
            ]
            prompt = self.tokenizer.apply_chat_template(
                tokenizer_chat_conversation, tokenize=False, add_generation_prompt=True
            )
            assert isinstance(prompt, Text)
            if not prompt:
                raise ValueError("The `prompt` cannot be empty, no content in messages")

        # Chat completion request
        completion_res = self.text_completion(
            prompt=prompt, model=model, stop=stop, **kwargs
        )
        # Parse completion response to chat completion
        chat_completion = ChatCompletion(
            id=completion_res.id,
            choices=[],
            created=completion_res.created,
            model=completion_res.model,
            object="chat.completion",
            system_fingerprint=completion_res.system_fingerprint,
            usage=completion_res.usage,
        )
        # Modify completion response to chat completion
        for c in completion_res.choices:
            message_text = c.text
            for stop_word in stop:
                modified_message_text = replace_right(
                    message_text, old=stop_word, new="", occurrence=1
                )
                if modified_message_text != message_text:
                    message_text = modified_message_text
                    break
            chat_completion.choices.append(
                ChatChoice(
                    finish_reason=c.finish_reason,
                    index=c.index,
                    message=ChatCompletionMessage(
                        role="assistant",
                        content=message_text,
                    ),
                )
            )
        return chat_completion

    def text_completion(
        self, prompt: Text, *args, model: Text, **kwargs
    ) -> "Completion":
        # Validate parameters
        if not prompt:
            raise ValueError("The `prompt` cannot be empty")
        if model != self.model_name:
            logger.warning(
                f"The model `{model}` is not the same as the action's model "
                + f"`{self.model_name}`"
            )

        # Initialize completion response
        completion_res = Completion(
            id=str(uuid.uuid4()),
            choices=[],
            created=int(time.time()),
            model=self.model_name,
            object="text_completion",
        )

        # Prepare kwargs
        max_length = kwargs.get("max_tokens") or kwargs.get("max_length") or None
        if max_length is not None:
            kwargs["max_length"] = int(max_length)
        else:
            kwargs["max_length"] = max_length = self.default_max_tokens
        kwargs.pop("max_tokens", None)
        kwargs.pop("best_of", None)  # TODO: Implement best_of
        kwargs.pop("echo", None)  # TODO: Implement echo
        kwargs.pop("n", None)  # TODO: Implement n
        kwargs.pop("frequency_penalty", None)  # TODO: Implement frequency_penalty
        kwargs.pop("presence_penalty", None)  # TODO: Implement presence_penalty
        kwargs.pop("stream", None)  # TODO: Implement stream
        kwargs.pop("logprobs", None)  # TODO: Implement logprobs
        total_stop_words: Sequence[Text] = must_list(self.stop_words) + must_list(
            kwargs.pop("stop", ())
        )

        # Stopping criteria
        stop_criteria: Optional["StoppingCriteria"] = None
        if len(total_stop_words) > 0:
            stop_criteria = StopAtWordsStoppingCriteria.from_stop_words(
                total_stop_words, self.tokenizer
            )
            kwargs["stopping_criteria"] = StoppingCriteriaList([stop_criteria])

        # Tokenize prompt
        inputs = self.tokenizer(
            prompt, return_tensors="pt", return_attention_mask=False
        )
        input_ids = inputs["input_ids"]
        input_ids = cast(torch.Tensor, input_ids)
        input_ids = input_ids.to(self.device)
        inputs_tokens_length = int(input_ids.shape[1])
        if max_length is not None and inputs_tokens_length >= max_length:
            logger.warning(
                "The input tokens length is already greater than max_length, "
                + f"{inputs_tokens_length} >= {max_length}, "
                + f"resetting max_length to {inputs_tokens_length + 20}"
            )
            kwargs["max_length"] = max_length = inputs_tokens_length + 20

        # Generate text completion
        outputs = self.model.generate(input_ids, **kwargs)
        outputs = cast(torch.Tensor, outputs)
        outputs_tokens_length = outputs.shape[1]
        completed_text = self.tokenizer.batch_decode(outputs)[0]
        output_text = self.remove_echo_text(
            prompt=prompt, completed_text=completed_text, tokenizer=self.tokenizer
        )

        # Collect completion response
        finish_reason = "length"
        if stop_criteria is not None:
            finish_reason = stop_criteria.get_stop_reason() or finish_reason
        elif (
            self.tokenizer.eos_token_id is not None
            and outputs[-1][-1].item() == self.tokenizer.eos_token_id
        ):
            finish_reason = "stop"
        completion_choice = CompletionChoice(
            finish_reason=finish_reason, index=0, text=output_text
        )
        completion_res.choices.append(completion_choice)
        completion_res.usage = CompletionUsage(
            total_tokens=outputs_tokens_length,
            prompt_tokens=inputs_tokens_length,
            completion_tokens=outputs_tokens_length - inputs_tokens_length,
        )
        completion_res.created = int(time.time())
        return completion_res

    def embeddings(
        self,
        input: Union[Text, PreTokenizedInput, List[PreTokenizedInput]],
        *args,
        model: Text,
        **kwargs,
    ) -> "CreateEmbeddingResponse":
        # Validate parameters
        kwargs.pop("encoding_format", None)  # TODO: Implement encoding_format
        kwargs.pop("output_hidden_states", None)

        # Tokenize prompt
        inputs = self.tokenizer(
            input, return_tensors="pt", padding=True, truncation=True
        )
        inputs = inputs.to(self.device)
        input_ids = self.ensure_tensor(inputs["input_ids"])
        inputs_tokens_length = int(input_ids.shape[1])

        with torch.no_grad():  # No need to compute gradients
            output = self.model(**inputs, output_hidden_states=True, **kwargs)
            hidden_states = output.hidden_states[-1]

        # Perform pooling.
        embeddings = mean_pooling(
            tensor_to_np(tensor=self.ensure_tensor(hidden_states)),
            tensor_to_np(self.ensure_tensor(inputs["attention_mask"])),
        )
        return CreateEmbeddingResponse.model_validate(
            {
                "data": [
                    {
                        "embedding": emb,
                        "index": idx,
                        "object": "embedding",
                    }
                    for idx, emb in enumerate(embeddings.tolist())
                ],
                "model": self.model_name,
                "object": "list",
                "usage": {
                    "total_tokens": inputs_tokens_length,
                    "prompt_tokens": inputs_tokens_length,
                },
            }
        )

    def ensure_tensor(self, input: Any) -> "torch.Tensor":
        if isinstance(input, torch.Tensor) is False:
            logger.warning(
                f"Input is not a tensor, converting to tensor: {type(input)}"
            )
        return input

    def read_model_name(self, **kwargs) -> Text:
        model_name = should_str_or_none(kwargs.get("model_name")) or self.MODEL_NAME
        if not model_name:
            raise ValueError(
                "The `model_name` cannot be empty. Please set the `HF_MODEL_NAME` "
                + "or `MODEL_NAME` environment variable or pass the `model_name` "
                + "parameter. Or you might want to override the "
                + f"{self.__class__.__name__}.MODEL_NAME attribute."
            )
        logger.info(f"Using model: {model_name}")
        return model_name

    def load_model_and_tokenizer(
        self, **kwargs
    ) -> Tuple[PreTrainedModel, Union[PreTrainedTokenizer, PreTrainedTokenizerFast]]:
        model_config = AutoConfig.from_pretrained(self.model_name)
        params = {
            "pretrained_model_name_or_path": self.model_name,
            "device_map": self.device,
            "trust_remote_code": True,
        }
        logger.debug(f"Loading model... : {self.model_name}")
        # Define flash attention config
        if (
            "use_flash_attention" in kwargs and kwargs["use_flash_attention"]
        ) or self.use_flash_attention is True:
            params["attn_implementation"] = "flash_attention_2"
            logger.info("Using Flash Attention 2.")
        # Define quantization config
        else:
            quantization_config = self.load_quantization_config(**kwargs)
            if quantization_config is not None:
                params["quantization_config"] = quantization_config
        # Define torch dtype
        if (
            kwargs.get("torch_dtype") is not None
            or self.dtype is not None
            or hasattr(model_config, "torch_dtype")
        ):
            if params.get("quantization_config"):
                logger.info(
                    "The `torch_dtype` is not used when `quantization_config` is set."
                )
            elif kwargs.get("torch_dtype") is not None:
                params["torch_dtype"] = self.dtype = kwargs["torch_dtype"]
            elif self.dtype is not None:
                params["torch_dtype"] = self.dtype
            elif hasattr(model_config, "torch_dtype"):
                params["torch_dtype"] = self.dtype = model_config.torch_dtype
        logger.info(f"Loading model with params: {params}")
        # Load model and tokenizer
        # Load causal language model
        if self.is_causal_lm is True:
            model = AutoModelForCausalLM.from_pretrained(**params)
        # Load other models
        else:
            model = AutoModel.from_pretrained(**params)
        tokenizer = AutoTokenizer.from_pretrained(
            self.model_name, trust_remote_code=True
        )
        return (model, tokenizer)

    def load_quantization_config(self, **kwargs) -> Optional["BitsAndBytesConfig"]:
        use_quantization = bool(
            kwargs["use_quantization"]
            if "use_quantization" in kwargs
            else self.use_quantization
        )
        if use_quantization is True:
            params = {}
            load_in_8bit = (
                kwargs["load_in_8bit"]
                if "load_in_8bit" in kwargs
                else self.load_in_8bit
            )
            load_in_4bit = (
                kwargs["load_in_4bit"]
                if "load_in_4bit" in kwargs
                else self.load_in_4bit
            )
            llm_int8_threshold = (
                kwargs["llm_int8_threshold"]
                if "llm_int8_threshold" in kwargs
                else self.llm_int8_threshold
            )
            llm_int8_skip_modules = (
                kwargs["llm_int8_skip_modules"]
                if "llm_int8_skip_modules" in kwargs
                else self.llm_int8_skip_modules
            )
            llm_int8_enable_fp32_cpu_offload = (
                kwargs["llm_int8_enable_fp32_cpu_offload"]
                if "llm_int8_enable_fp32_cpu_offload" in kwargs
                else self.llm_int8_enable_fp32_cpu_offload
            )
            llm_int8_has_fp16_weight = (
                kwargs["llm_int8_has_fp16_weight"]
                if "llm_int8_has_fp16_weight" in kwargs
                else self.llm_int8_has_fp16_weight
            )
            if load_in_8bit is not None:
                params["load_in_8bit"] = load_in_8bit
            if load_in_4bit is not None:
                params["load_in_4bit"] = load_in_4bit
                if load_in_4bit is True:
                    params["bnb_4bit_compute_dtype"] = torch.float16
            if llm_int8_threshold is not None:
                params["llm_int8_threshold"] = llm_int8_threshold
            if llm_int8_skip_modules is not None:
                params["llm_int8_skip_modules"] = llm_int8_skip_modules
            if llm_int8_enable_fp32_cpu_offload is not None:
                params["llm_int8_enable_fp32_cpu_offload"] = (
                    llm_int8_enable_fp32_cpu_offload
                )
            if llm_int8_has_fp16_weight is not None:
                params["llm_int8_has_fp16_weight"] = llm_int8_has_fp16_weight

            logger.info(f"Using quantization config: {params}")
            return BitsAndBytesConfig(**params)
        else:
            return None

    def remove_echo_text(
        self,
        prompt: Text,
        completed_text: Text,
        tokenizer: Union[PreTrainedTokenizer, PreTrainedTokenizerFast],
    ) -> Text:
        """Remove the echo text from the completed text.

        Parameters
        ----------
        prompt : Text
            The prompt text.
        completed_text : Text
            The completed text.
        tokenizer : Union[PreTrainedTokenizer, PreTrainedTokenizerFast]
            The tokenizer used to tokenize the prompt and completed text.

        Returns
        -------
        Text
            The output text without the echo text.
        """

        if tokenizer.eos_token:
            _pat = rf"\s*{re.escape(tokenizer.eos_token)}\s*"
            prompt = re.sub(_pat, "", prompt)
            completed_text = re.sub(_pat, "", completed_text)
        if tokenizer.bos_token:
            _pat = rf"\s*{re.escape(tokenizer.bos_token)}\s*"
            prompt = re.sub(_pat, "", prompt)
            completed_text = re.sub(_pat, "", completed_text)
        output_text = completed_text.replace(prompt, "", 1)
        output_text = remove_special_tokens(output_text, tokenizer=self.tokenizer)
        return output_text



import importlib
import time
from typing import TYPE_CHECKING
from unittest.mock import MagicMock, patch

import pytest
from fastapi.testclient import TestClient

import languru.server.main
from languru.server.config import AppType

if TYPE_CHECKING:
    from _pytest.monkeypatch import MonkeyPatch


@pytest.fixture
def llm_env(monkeypatch: "MonkeyPatch"):
    monkeypatch.setenv("APP_TYPE", AppType.llm)


@pytest.fixture
def agent_env(monkeypatch: "MonkeyPatch"):
    monkeypatch.setenv("APP_TYPE", AppType.agent)


@pytest.fixture
def mocked_model_discovery_list():
    from languru.examples.return_values._openai import return_model
    from languru.resources.model.discovery import ModelDiscovery, SqlModelDiscovery

    with patch.object(
        ModelDiscovery, "list", MagicMock(return_value=[return_model])
    ), patch.object(SqlModelDiscovery, "list", MagicMock(return_value=[return_model])):
        yield


@pytest.fixture
def mocked_model_discovery_retrieve():
    from languru.examples.return_values._openai import return_model
    from languru.resources.model.discovery import ModelDiscovery, SqlModelDiscovery

    with patch.object(
        ModelDiscovery, "retrieve", MagicMock(return_value=return_model)
    ), patch.object(
        SqlModelDiscovery, "retrieve", MagicMock(return_value=return_model)
    ):
        yield


def test_llm_app_models(llm_env):
    importlib.reload(languru.server.main)

    with TestClient(languru.server.main.app) as client:
        response = client.get("/v1/models")
        assert response.status_code == 200
        data = response.json()
        assert data["object"] == "list"
        assert len(data["data"]) > 0


def test_llm_app_model_retrieve(llm_env):
    importlib.reload(languru.server.main)

    with TestClient(languru.server.main.app) as client:
        response = client.get("/v1/models/gpt-3.5-turbo")
        assert response.status_code == 200

        response = client.get("/v1/models/MODEL_NOT_FOUND")
        assert response.status_code == 404


def test_llm_app_model_register(llm_env):
    importlib.reload(languru.server.main)

    with TestClient(languru.server.main.app) as client:
        model = {
            "id": "model_id",
            "created": int(time.time()),
            "object": "model",
            "owned_by": "pytest",
        }
        response = client.post("/v1/models/register", json=model)
        assert response.status_code == 403


def test_agent_app_models(agent_env, mocked_model_discovery_list):
    importlib.reload(languru.server.main)

    with TestClient(languru.server.main.app) as client:
        response = client.get("/v1/models")
        assert response.status_code == 200
        data = response.json()
        assert data["object"] == "list"
        assert len(data["data"]) > 0


def test_agent_app_model_retrieve(agent_env, mocked_model_discovery_retrieve):
    importlib.reload(languru.server.main)

    with TestClient(languru.server.main.app) as client:
        response = client.get("/v1/models/model_id")
        assert response.status_code == 200


def test_agent_app_model_register(agent_env):
    importlib.reload(languru.server.main)

    with TestClient(languru.server.main.app) as client:
        model = {
            "id": "test_model_id",
            "created": int(time.time()),
            "object": "model",
            "owned_by": "pytest",
        }
        response = client.post("/v1/models/register", json=model)
        assert response.status_code == 200



import json
from typing import Any, List, Optional, Text, Tuple, TypeVar

from pydantic import BaseModel
from rich import box
from rich import print as rich_print
from rich.style import StyleType
from rich.table import Table
from rich.text import Text as RichText

from languru.config import logger

T = TypeVar("T")


def should_str_or_none(value: Text | Any) -> Optional[Text]:
    if isinstance(value, Text):
        return value
    elif value is None:
        return None
    logger.warning(f"Value {value} is not a string, returning None")
    return None


def should_str(value: Text | Any) -> Text:
    if should_str_or_none(value) is None:
        raise ValueError(f"Value {value} is not a string")
    return value


def must_list_or_none(
    value: List[T] | Any, return_none_if_empty: bool = False
) -> Optional[List[T]]:
    if isinstance(value, List):
        if return_none_if_empty and len(value) == 0:
            return None
        return value
    elif isinstance(value, Tuple):
        if return_none_if_empty and len(value) == 0:
            return None
        return list(value)
    elif value is None:
        return None
    else:
        return [value]


def must_list(value: List[T] | Any) -> List[T]:
    to_items = must_list_or_none(value)
    if to_items is None:
        raise ValueError(f"Could not convert {value} to a list")
    return to_items


def debug_print(
    *values: Any,
    title: Text = "Debug Print",
    box: box.Box | None = box.HEAVY_HEAD,
    colors: List[StyleType] = [
        "bright_blue",
        "bright_cyan",
        "bright_green",
        "bright_magenta",
    ],
) -> None:
    if not values:
        return
    tb = Table(title=RichText(title), box=box, show_header=False)
    for idx, value in enumerate(values):
        style = colors[idx % len(colors)]
        if isinstance(value, BaseModel):
            tb.add_row(value.model_dump_json(indent=2), style=style)
        elif isinstance(value, dict):
            tb.add_row(json.dumps(value, indent=2, ensure_ascii=False), style=style)
        else:
            tb.add_row(str(value), style=style)
    rich_print(tb)


def replace_right(source_str: Text, old: Text, new: Text, occurrence: int = -1) -> Text:
    return source_str[::-1].replace(old[::-1], new[::-1], occurrence)[::-1]


def str_strong_casefold(text: Text) -> Text:
    return text.strip().replace("-_. ", "").casefold()



import socket
from typing import Text


def check_port(port: int, host: Text = "0.0.0.0"):
    """Check if a port is available.

    Parameters
    ----------
    port : int
        The port number to check.
    host : str, optional
        The host address to bind to. Defaults to '0.0.0.0'.

    Returns
    -------
    bool
        True if the port is available, False otherwise.
    """

    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        try:
            s.bind((host, port))
            return True
        except socket.error:
            return False


# Get available port starting from a port number and incrementing by 1
def get_available_port(start_port: int, host: Text = "0.0.0.0") -> int:
    """Get an available port starting from a port number and incrementing by 1.

    Parameters
    ----------
    start_port : int
        The port number to start from.
    host : str, optional
        The host address to bind to. Defaults to '0.0.0.0'.

    Returns
    -------
    int
        The available port number.
    """

    port = start_port
    while not check_port(port=port, host=host):
        port += 1
    return port
